#!/usr/bin/python3
"""Test suite for the database abstraction module (model)."""

# Too many public methods
# pylint: disable=R0904

# Standard modules
import unittest

import sqlalchemy
from sqlalchemy import (Column, ForeignKey, Integer, MetaData, String, Table,
                        create_engine)
from sqlalchemy.exc import IntegrityError, OperationalError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import lazyload, relationship

import uweb3
from uweb3.alchemy_model import AlchemyRecord
from uweb3.libs.sqltalk import mysql

# ##############################################################################
# Record classes for testing
#
Base = declarative_base()

class BasicTestRecord(AlchemyRecord, Base):
  """Test record for offline tests."""
  __tablename__ = 'basicTestRecord'
  ID = Column(Integer, primary_key=True)
  name = Column(String(32), nullable=False)
  x = Column(String(32))

class Author(AlchemyRecord, Base):
  __tablename__ = 'author'
  ID = Column(Integer, primary_key=True)
  name = Column(String(32), nullable=False)

class Writers(AlchemyRecord, Base):
  __tablename__ = 'writers'
  ID = Column(Integer, primary_key=True)
  name = Column(String(32), nullable=False)

class Book(AlchemyRecord, Base):
  """Book class for testing purposes."""
  __tablename__ = 'book'
  ID = Column(Integer, primary_key=True)
  title = Column(String(32), nullable=False)
  authorid = Column('authorid', Integer, ForeignKey('author.ID'))
  author = relationship("Author",  lazy="select")

class BaseRecordTests(unittest.TestCase):
  """Offline tests of methods and behavior of the BaseRecord class."""
  def setUp(self):
    """Sets up the tests for the offline Record test."""
    self.record_class = BasicTestRecord

  def testTableName(self):
    """[BaseRecord] TableName returns the expected value and obeys _TABLE"""
    self.assertEqual(self.record_class.TableName(), 'basicTestRecord')

  def testPrimaryKey(self):
    """[BaseRecord] Primary key value on `key` property, default field 'ID'"""
    record = self.record_class(None, {'ID': 12, 'name': 'J.R.R. Tolkien'})
    self.assertEqual(record.key, 12)

  def testEquality(self):
    """[BaseRecord] Records of the same content are equal to eachother"""
    record_one = self.record_class(None, {'ID': 2, 'name': 'Rowling'})
    record_two = self.record_class(None, {'ID': 2, 'name': 'Rowling'})
    record_three = self.record_class(None, {'ID': 3, 'name': 'Rowling'})
    record_four = self.record_class(None, {'ID': 2, 'name': 'Rowling', 'x': 2})
    self.assertFalse(record_one is record_two)
    self.assertEqual(record_one, record_two)
    self.assertNotEqual(record_one, record_three)
    self.assertNotEqual(record_one, record_four)

class RecordTests(unittest.TestCase):
  """Online tests of methods and behavior of the Record class."""

  def setUp(self):
    """Sets up the tests for the Record class."""
    self.engine = DatabaseConnection()
    self.session = Create_session(self.engine)
    self.meta = MetaData()
    book = Table(
      'book', self.meta,
      Column('ID', Integer, primary_key=True),
      Column('authorid', Integer, ForeignKey('author.ID')),
      Column('title', String(32), nullable=False)
    )
    author = Table(
      'author', self.meta,
      Column('ID', Integer, primary_key=True),
      Column('name', String(32), nullable=False),
    )
    writers = Table(
      'writers', self.meta,
      Column('ID', Integer, primary_key=True),
      Column('name', String(32), nullable=False),
    )
    self.meta.create_all(self.engine)

  def tearDown(self):
    """Destroy tables after testing."""
    self.session.close()
    Book.__table__.drop(self.engine)
    Author.__table__.drop(self.engine)

  def testLoadPrimary(self):
    """[Record] Records can be loaded by primary key using FromPrimary()"""
    inserted = Author.Create(self.session, {'name': 'A. Chrstie'})
    author = Author.FromPrimary(self.session, inserted.key)
    self.assertEqual(type(author), Author)
    self.assertEqual(len(author), 2)
    self.assertEqual(author.key, author.ID)
    self.assertEqual(author.name, 'A. Chrstie')

  def testCreateRecord(self):
    """Database records can be created using Create()"""
    new_author = Author.Create(self.session, {'name': 'W. Shakespeare'})
    author = Author.FromPrimary(self.session, new_author.key)
    self.assertEqual(author.name, 'W. Shakespeare')

  def testCreateRecordWithBadField(self):
      """Database record creation fails if there are unknown fields present"""
      self.assertRaises(AttributeError, Author.Create, self.session,
                        {'name': 'L. Tolstoy', 'email': 'leo@tolstoy.ru'})

  def testUpdateRecord(self):
    """The record can be given new values and these are properly stored"""
    author = Author.Create(self.session, {'name': 'B. King'})
    author.name = 'S. King'
    author.Save()
    same_author = Author.FromPrimary(self.session, author.key)
    self.assertEqual(author.name, 'S. King')
    self.assertEqual(author, same_author)

  def testUpdatingDeletedRecord(self):
    """Should raise an error because the record no longer exists"""
    author = Author.Create(self.session, {'name': 'B. King'})
    Author.DeletePrimary(self.session, author.key)
    author.name = 'S. King'
    self.assertRaises(sqlalchemy.orm.exc.StaleDataError, author.Save())

  def testUpdatePrimaryKey(self):
    """Saving with an updated primary key properly saved the record"""
    author = Author.Create(self.session, {'name': 'C. Dickens'})
    self.assertEqual(author.key, 1)
    author.ID = 101
    author.Save()
    self.assertRaises(uweb3.model.NotExistError, Author.FromPrimary,
                      self.session, 1)
    same_author = Author.FromPrimary(self.session, 101)
    self.assertEqual(same_author, author)

  def testLoadRelated(self):
    """Fieldnames that match tablenames trigger automatic loading"""
    author = Author.Create(self.session, {'name': 'D. Koontz'})
    book = Book.Create(self.session, {'title': 'The eyes of Darkness', 'authorid': 1})
    self.assertEqual(type(author), Author)
    self.assertEqual(type(book.author), Author)
    self.assertEqual(book.author.name, 'D. Koontz')
    self.assertEqual(book.author.key, 1)

  def testLoadRelatedFailure(self):
    """Automatic loading raises IntegrityError if the foreign record is absent"""
    self.assertRaises(IntegrityError, Book.Create, self.session, {'title': 'The eyes of Darkness', 'authorid': 1})

  def testLoadRelatedSuppressedForNone(self):
    """Automatic loading is not attempted when the field value is `None`"""
    self.assertRaises(OperationalError, Book.Create, self.session, {'title': None})

  def testVerifyNoLoad(self):
    """No loading is performed on a field that matches a class but no table"""
    self.assertRaises(AttributeError, Book, self.session, {'writer': 1})

  def testValues(self):
    author = Author.Create(self.session, {'name': 'D. Koontz'})
    self.assertEqual(author.values(), [1, 'D. Koontz'])
    book = Book.Create(self.session, {'title': 'The eyes of Darkness', 'authorid': author.key})
    self.assertEqual(book.values(), [1, 'The eyes of Darkness', 1, Author(None, {'ID': 1, 'name': 'D. Koontz'})])

  def testItems(self):
    author = Author.Create(self.session, {'name': 'D. Koontz'})
    self.assertEqual(author.items(), [('ID', 1), ('name', 'D. Koontz')])
    book = Book.Create(self.session, {'title': 'The eyes of Darkness', 'authorid': author.key})
    self.assertEqual(book.items(),
                     [('ID', 1),
                      ('title', 'The eyes of Darkness'),
                      ('authorid', 1),
                      ('author', Author(None, {'ID': 1, 'name': 'D. Koontz'}))])

def DatabaseConnection():
  """Returns an SQLTalk database connection to 'uWeb3_model_test'."""
  return create_engine('mysql://{user}:{passwd}@{host}/{db}'.format(
    host='localhost',
    user='stef',
    passwd='24192419',
    db='uweb_test'
  ))

def Create_session(engine):
  from sqlalchemy.orm import sessionmaker
  Session = sessionmaker(autocommit=False)
  Session.configure(bind=engine)
  return Session()

if __name__ == '__main__':
  unittest.main(testRunner=unittest.TextTestRunner(verbosity=2))



#!/usr/bin/python3
"""Easy use SQL abstraction module.

Returns a custom QueryResult object that holds
the result, query, used character set and various other small statistics. The
QueryResult object also support pivoting and subselects

Currently implements a MySQL and sqlite abstraction module with a stripped down
version of MySQLdb internally.

example usage:

  from sqltalk import mysql
  connection = mysql.Connect()
  with connection as cursor:
    result = cursor.RawQuery('select * from table')
"""



#!/usr/bin/python3
"""SQLTalk SQLite Cursor class."""
__author__ = 'Elmer de Looff <elmer@underdark.nl>'
__version__ = '0.4'


# Custom modules
from .. import sqlresult


class Cursor:
  def __init__(self, connection):
    self.connection = connection
    self.cursor = connection.cursor()

  def Execute(self, query, args=(), many=False):
    try:
      if many:
        result = self.cursor.executemany(query, args)
      else:
        result = self.cursor.execute(query, args)
    except Exception:
      self.connection.logger.exception('Exception during query execution')
      raise
    fieldnames = [field[0] for field in result.description]
    return sqlresult.ResultSet(
        affected=result.rowcount,
        charset='utf-8',
        fields=fieldnames,
        insertid=result.lastrowid,
        query=(query, tuple(args)),
        result=[dict(zip(fieldnames, row)) for row in result.fetchall()],
    )

  def Insert(self, table, values):
    if not values:
      raise ValueError('Must insert 1 or more value')
    elif isinstance(values, dict):
      query = ('INSERT INTO %s (%s) VALUES (%s)' %
               (table,
                ', '.join(map(self.connection.EscapeField, values)),
                ', '.join('?' * len(values))))
      return self.Execute(query, args=values.values(), many=False)
    query = ('INSERT INTO %s (%s) VALUES (%s)' %
             (table,
              ', '.join(map(self.connection.EscapeField, values[0])),
              ', '.join('?' * len(values[0]))))
    return self.Execute(
        query, args=(row.values() for row in values), many=True)

  def Select(self, table, fields=None, conditions=None, order=None, group=None,
             limit=None, offset=0):
    """Select fields from table that match the conditions, ordered and limited.

    Arguments:
      table:      string/list/tuple. Table(s) to select fields out of.
      fields:     string/list/tuple (optional). Fields to select. Default '*'.
                  As string, single field name. (autoquoted)
                  As list/tuple, one field name per element. (autoquoted)
      conditions: string/list/tuple (optional). SQL 'where' statement.
                  Literal as string. AND'd if list/tuple.
                  THESE WILL NOT BE ESCAPED FOR YOU, EVER.
      order:      (nested) list/tuple (optional).
                  Defines sorting of table before updating, elements can be:
                    string: a field to order by (in default database order).
                    list/tuple of two elements:
                      1) string, field name to order by
                      2) bool, revserse; set this to True to reverse the order
      group:      str (optional). Field name or function to group result by.
      limit:      integer (optional). Defines output size in rows.
      offset:     integer (optional). Number of rows to skip, requires limit.

    Returns:
      sqlresult.ResultSet object.
    """
    if isinstance(table, str):
      table = self.connection.EscapeField(table)
    else:
      table = ', '.join(map(self.connection.EscapeField, table))

    if fields is None:
      fields = '*'
    elif isinstance(fields, str):
      fields = self.connection.EscapeField(fields)
    else:
      fields = ', '.join(map(self.connection.EscapeField, fields))

    #FIXME(Elmer): Add consistent programmatic condition support for SQLTalk.
    if isinstance(conditions, (list, tuple)):
      conditions = ' AND '.join(conditions)
    elif not conditions:
      conditions = 1

    if order is not None:
      orders = []
      for rule in order:
        if isinstance(rule, str):
          orders.append(self.connection.EscapeField(rule))
        else:
          orders.append('%s %s' %
                        (self.connection.EscapeField(rule[0]),
                        ('ASC', 'DESC')[rule[1]]))
      order = 'ORDER BY ' + ', '.join(orders)
    else:
      order = ''

    if group is not None:
      group = 'GROUP BY ' + self.connection.EscapeField(group)
    else:
      group = ''

    if limit is not None:
      if offset:
        limit = 'LIMIT %d OFFSET %d' % (limit, offset)
      else:
        limit = 'LIMIT %d' % limit
    else:
      limit = ''

    query = ('SELECT %s FROM %s WHERE %s %s %s %s' %
             (fields, table, conditions, group, order, limit))
    return self.Execute(query)



#!/usr/bin/python3
"""This module implements the Connection class, which sets up a connection to
an SQLite database. From this connection, cursor objects can be created, which
use the escaping and character encoding facilities offered by the connection.
"""
__author__ = 'Elmer de Looff <elmer@underdark.nl>'
__version__ = '0.3'

# Standard modules
import sqlite3
import logging
import os
import queue
import threading

# Application specific modules
from . import converters
from . import cursor

COMMIT = '----COMMIT'
ROLLBACK = '----ROLLBACK'
NAMED_TYPE_SELECT = 'SELECT `name` FROM `sqlite_master` where `type`=?'

class Connection(sqlite3.Connection):
  def __init__(self, *args, **kwds):
    db_name = os.path.splitext(os.path.split(args[0])[1])[0]
    self.logger = logging.getLogger('sqlite_%s' % db_name)
    if kwds.pop('debug', False):
      self.logger.setLevel(logging.DEBUG)
    else:
      self.logger.setLevel(logging.WARNING)
    if kwds.pop('disable_log', False):
      self.logger.disable_logger = True
    sqlite3.Connection.__init__(self, *args, **kwds)

  def __enter__(self):
    """Starts a transaction."""
    self.logger.debug('Beginning new transaction.')
    return cursor.Cursor(self)

  def __exit__(self, exc_type, _exc_value, _exc_traceback):
    """End of transaction: commits , or rolls back on failure."""
    if exc_type:
      self.rollback()
      self.logger.warning('Transaction was rolled back.')
    else:
      self.commit()
      self.logger.debug('Transaction committed.')

  def commit(self):
    sqlite3.Connection.commit(self)

  def rollback(self):
    sqlite3.Connection.rollback(self)

  @staticmethod
  def EscapeField(field):
    """Returns a SQL escaped field or table name."""
    return '.'.join('`%s`' % f.replace('`', '``') for f in field.split('.'))

  def EscapeValues(self, obj):
    """We do not escape here, we simple return the value and allow the query
    engine to escape using parameters.
    """
    return obj

  def ShowTables(self):
    result = self.execute(NAMED_TYPE_SELECT, ('table',)).fetchall()
    return [row[0] for row in result]


class ThreadedConnection(threading.Thread):
  def __init__(self, *args, **kwds):
    super(ThreadedConnection, self).__init__()
    # Set up a logger
    db_name = os.path.splitext(os.path.split(args[0])[1])[0]
    self.logger = logging.getLogger('sqlite_%s' % db_name)
    if kwds.pop('debug', False):
      self.logger.setLevel(logging.DEBUG)
    else:
      self.logger.setLevel(logging.WARNING)
    if kwds.pop('disable_log', False):
      self.logger.disable_logger = True

    self.sqlite_args = args
    self.sqlite_kwds = kwds
    self.queries = queue.queue(1)
    self.transaction_lock = threading.RLock()
    self.daemon = True
    self.start()

  def __enter__(self):
    """Starts a transaction."""
    self.transaction_lock.acquire()
    return cursor.Cursor(self)

  def __exit__(self, exc_type, _exc_value, _exc_traceback):
    """End of transaction: commits, or rolls back on failure."""
    if exc_type:
      self.rollback()
      self.logger.warning('Transaction was rolled back.')
    else:
      self.commit()
      self.logger.debug('Transaction committed.')
    self.transaction_lock.release()

  def commit(self):
    self.execute(COMMIT)

  def execute(self, query, args=()):
    with self.transaction_lock:
      response = queue.queue()
      self.queries.put((query, args, response, False))
      return self._ProcessResponse(response)

  def executemany(self, query, args=()):
    with self.transaction_lock:
      response = queue.queue()
      self.queries.put((query, args, response, True))
      return self._ProcessResponse(response)

  def rollback(self):
    self.execute(ROLLBACK)

  def run(self):
    connection = Connection(*self.sqlite_args, **self.sqlite_kwds)
    while True:
      query, args, response, many = self.queries.get()
      try:
        if query is COMMIT:
          response.put(connection.commit())
        elif query is ROLLBACK:
          response.put(connection.rollback())
        else:
          execute = connection.executemany if many else connection.execute
          result = execute(query, args)
          response.put(SqliteResult(result.fetchall(), result.description,
                                    result.rowcount, result.lastrowid))
          del execute, result
      except Exception as error:
        response.put(error)
        del error

  @staticmethod
  def _ProcessResponse(response):
    """Processes the response given by the SQLite connection thread.

    If the response is an exception type, it will be raised in this thread, so
    that the underlying thread doesn't silently fail. This allows a database
    error to be handled by the calling code properly.
    """
    response = response.get()
    if isinstance(response, Exception):
      raise response
    return response

  @staticmethod
  def EscapeField(field):
    """Returns a SQL escaped field or table name."""
    return '.'.join('`%s`' % f.replace('`', '``') for f in field.split('.'))

  def ShowTables(self):
    result = self.execute(NAMED_TYPE_SELECT, ('table',)).fetchall()
    return [row[0] for row in result]


class SqliteResult:
  def __init__(self, result, description, rowcount, lastrowid):
    self.result = result
    self.description = description
    self.rowcount = rowcount
    self.lastrowid = lastrowid

  def fetchall(self):
    return self.result


#FIXME(Elmer): This needs defining in one place, not in each and every file.
DataError = sqlite3.DataError
DatabaseError = sqlite3.DatabaseError
Error = sqlite3.Error
IntegrityError = sqlite3.IntegrityError
InterfaceError = sqlite3.InterfaceError
InternalError = sqlite3.InternalError
NotSupportedError = sqlite3.NotSupportedError
OperationalError = sqlite3.OperationalError



#!/usr/bin/python3
"""This module contains a set of classes which can be used in a multi-facet
environment like a web-application to maintain knowledge about the origin of
strings. It also functions as a system that allows for mixing strings of various
origins and to make them safe for the requested enviroment.

Currently the goal is to verify that strings never create unintended issues like
cross site scripting, json injection and similar.

The premise is that all strings coming from the web-user, eg form inputs,
headers, remote api's, database fields and config files are unsafe by definition
 and that all consumers of strings upgrade to their respective safe string
 variant at ingestion.

Consumers like template parsers know what the intended working environment for a
 string is going to be, eg, html. And thus can decide on escaping if the handed
 object is not of a type known to be safe for that environment.

Additions of unsafe types to a safe type automatically escape up to the safe
type. Handy escape() functions are present to do manual escaping if required.
"""

#TODO: logger, dont output Enters
# bash injection
# mysql escaping
__author__ = 'Jan Klopper (jan@underdark.nl)'
__version__ = 0.1

import html
from json import JSONEncoder
import json
import urllib.parse as urlparse
import re

# json encoder modules
import datetime
import uuid
from uweb3 import model


class Basesafestring(str):
  """Base safe string class
  This does not signal any safety against injection itself, use the child
  classes instead!
  """

  def __add__(self, other):
    """Adds a second string to this string, upgrading it in the process"""
    data = ''.join(( # do not use the __add__ since that creates a loop
        self, # the original item
        self.__upgrade__(other)))
    return self.__class__(data)

  def __upgrade__(self, other):
    """Upgrade a given object to be as safe, and in the same safety context as
    the current object"""
    if type(other) == self.__class__: #same type, easy, lets add
      return other
    if isinstance(other, Basesafestring): # lets unescape the other 'safe' type,
      otherdata = other.unescape(other) # its escaping is not needed for his context
      return self.escape(otherdata) # escape it using our context
    return self.escape(str(other)) # escape it using our context

  def __new__(cls, data, **kwargs):
    return super().__new__(cls,
        cls.escape(cls, str(data)) if 'unsafe' in kwargs else data)

  def __str__(self):
    if self.__class__ == Basesafestring:
      raise NotImplementedError
    return super().__str__()

  def __repr__(self):
    if self.__class__ == Basesafestring:
      raise NotImplementedError
    return super().__repr__()

  def format(self, *args, **kwargs):
    args = list(map(self.__upgrade__, args))
    kwargs = {k: self.__upgrade__(v) for k, v in kwargs.items()}
    return super().format(*args, **kwargs)

  def escape(self, data):
    raise NotImplementedError

  def unescape(self, data):
    raise NotImplementedError

  def join(self, items):
    output = []
    for item in items:
      output.append(self.__upgrade__(item))
    return self.__class__(''.join(output))


class SQLSAFE(Basesafestring):
  CHARS_ESCAPE_DICT = {
    '\0': '\\0',
    '\b': '\\b',
    '\t': '\\t',
    '\n': '\\n',
    '\r': '\\r',
    '\x1a': '\\Z',
    '"': '\\"',
    '\'': '\\\'',
    '\\': '\\\\'
  }

  CHARS_UNESCAPE_DICT = {
    '\\0': '\0',
    '\\b': '\b',
    '\\t': '\t',
    '\\n': '\n',
    '\\r': '\r',
    '\\Z': '\x1a',
    '\\"': '"',
    '\\\'': '\'',
    '\\\\': '\\'
  }

  CHARS_ESCAPE_REGEX = re.compile(r"""[\0\b\t\n\r\x1a\"\'\\]""")
  CHARS_UNESCAPE_REGEX = re.compile(r"""((\\t)|(\\0)|(\\n)|(\\b)|(\\r)|(\\Z)|(\\")|(\\\\')|(\\\\\\))""")
  PLACEHOLDERS_REGEX = re.compile(r"""\?+""")
  QUOTES_REGEX = re.compile(r"""([\"'])(?:(?=(\\?))\2.)*?\1""", re.DOTALL)

  def __upgrade__(self, other):
    """Upgrade a given object to be as safe, and in the same safety context as
    the current object
    """
    if type(other) == self.__class__: #same type, easy, lets add
      return other
    elif isinstance(other, Basesafestring): # lets unescape the other 'safe' type,
      return self.sanitize(other.unescape(other)) # escape it using our context
    return self.sanitize(" " + other, with_quotes=False)

  @classmethod
  def sanitize(cls, value, with_quotes=True):
    index = 0
    escaped = ""
    if len(cls.CHARS_ESCAPE_REGEX.findall(value)) == 0:
      if not str.isdigit(value):
        if with_quotes:
          return f"'{value}'"
        return value
      return value
    for m in cls.CHARS_ESCAPE_REGEX.finditer(value):
      escaped += value[index:m.span()[0]] + cls.CHARS_ESCAPE_DICT[m.group()]
      index = m.span()[1]
    escaped += value[index:]
    if not str.isdigit(escaped):
      if with_quotes:
        return f"'{escaped}'"
      return escaped
    return escaped

  def escape(self, sql, values):
    x = 0
    escaped = ""
    if not isinstance(values, tuple):
      raise ValueError("Values should be a tuple")

    if len(self.PLACEHOLDERS_REGEX.findall(sql)) != len(values):
      raise ValueError("Number of values does not match number of replacements")

    for index, m in enumerate(self.PLACEHOLDERS_REGEX.finditer(sql)):
      escaped += sql[x:m.span()[0]] + self.sanitize(values[index])
      x = m.span()[1]
    escaped += sql[x:]
    return SQLSAFE(escaped)

  def unescape(self, value):
    if not isinstance(value, SQLSAFE):
      raise ValueError(f"The value needs to be an instance of the SQLSAFE class and not of type: {type(value)}")
    x = 0
    escaped = ""
    for index, m in enumerate(self.CHARS_UNESCAPE_REGEX.finditer(value)):
      escaped += value[x:m.span()[0]]
      target = value[m.span()[0]:m.span()[1]]
      escaped += self.CHARS_UNESCAPE_DICT.get(target)
      x = m.span()[1]
    escaped += value[x:]
    return SQLSAFE(escaped)


# what follows are the actual useable classes that are safe in specific contexts
class Unsafestring(Basesafestring):
  """This class removes any escaping done"""

  def escape(self, data):
    return data

  def unescape(self, data):
    return data


class HTMLsafestring(Basesafestring):
  """This class signals that the content is HTML safe"""

  def escape(self, data):
    return html.escape(data)

  def unescape(self, data):
    return html.unescape(data)


class JSONsafestring(Basesafestring):
  """This class signals that the content is JSON safe

  Most of this will be handled by just feeding regular python objects into
  json.dumps, but for some outputs this might be handy. Eg, when outputting
  partial json into dynamic generated javascript files
  """

  def __new__(cls, data, **kwargs):
    if isinstance(data, str):
      data = cls.escape(cls, str(data)) if 'unsafe' in kwargs else data
    else:
      data = json.dumps(data, cls=JsonEncoder)
    return super().__new__(cls, data)

  def escape(self, data):
    return json.dumps(data, cls=JsonEncoder)

  def unescape(self, data):
    if not isinstance(data, str):
      raise TypeError
    return json.loads(data)

class JsonEncoder(json.JSONEncoder):
  def default (self, o):
    if isinstance(o, datetime.datetime):
      return o.strftime('%F %T')
    if isinstance(o, datetime.date):
      return o.strftime('%F')
    if isinstance(o, datetime.time):
      return o.strftime('%T')
    if isinstance(o, uuid.UUID):
      return str(o)
    if hasattr(o, "__json__"):
      return str(o.__json__())
    if hasattr(o, "__html__"):
      return str(o.__html__())
    if hasattr(o, "__dict__"):
      return o.__dict__
    try:
      return super().default(o)
    except TypeError:
      return str(o)


class URLqueryargumentsafestring(Basesafestring):
  """This class signals that the content is URL query argument safe"""

  def escape(self, data):
    """Encode any non url argument chars"""
    return urlparse.quote_plus(data)

  def unescape(self, data):
    """Decode any encoded non url argument chars"""
    return urlparse.unquote_plus(data)


class URLsafestring(Basesafestring):
  """This class signals that the content is URL safe, for use in http headers
  like redirects, but also calls to wget or the like
  """

  def escape(self, data):
    """Drops everything that does not fit in a url

    Since urlparse does not filter out new line chars well do that ourselves

    http://www.ietf.org/rfc/rfc1738.txt
    http://www.ietf.org/rfc/rfc2396.txt
    http://www.ietf.org/rfc/rfc3986.txt
    https://tools.ietf.org/html/rfc3986#section-2 uri chars
    """
    if '\n' in data or '\r' in data:
      data = data.splitlines()[0]
    return urlparse.urlparse(data).geturl()

  def unescape(self, data):
    """Can't unremove non url elements so we'll just return the string"""
    return data


class EmailAddresssafestring(Basesafestring):
  """This class signals that the content is a safe Email address

  Its usefull when sending out emails or constructing email headers
  Email Header injection is subverted.
  """

  def escape(self, data):
    """Drops everything that does not fit in an email address"""
    regex = re.compile(r'''(
        [a-zA-Z0-9._%+-]+ # username
        @ # @ symbol
        ([a-zA-Z0-9.-]+) # domain name
        (\.[a-zA-Z]{2,4}) # dot-something
    )''', re.VERBOSE)
    return regex.search(data).group()

  def unescape(self, data):
    """Can't unremove non address elements so we'll just return the string"""
    return data


class EmailHeadersafestring(Basesafestring):
  """This class signals that the content is a safe Email header

  Its usefull when sending out emails or constructing email headers."""

  def escape(self, data):
    """Drops everything that does not fit in a email header"""
    return data.replace("\n", "").replace("\r", "")

  def unescape(self, data):
    """Can't unremove non header elements so we'll just return the string"""
    return data



#!/usr/bin/python
"""SQLTalk MySQL Cursor class."""
__author__ = 'Elmer de Looff <elmer@underdark.nl>'
__version__ = '0.13'

# Standard modules
import warnings
import weakref
import pymysql

class ReturnObject(tuple):
  """An object that functions as a tuple but has more required attributes."""

  def __new__(cls, connection, results):
    """Creates the immutable tuple."""
    return super().__new__(cls, tuple(results))

  def __init__(self, connection, results):
    """Adds the required attributes."""
    if connection._result is not None:
      self.insertid = connection._result.insert_id
    else:
      self.insertid = None


class Cursor(pymysql.cursors.DictCursor):
  """Cursor to execute database interaction with, within a transaction."""

  def __init__(self, connection):
    self.description = None
    self.rownumber = 0
    self.rowcount = -1
    self.arraysize = 1
    self._executed = None
    self._result = None
    self._rows = None
    self._warnings_handled = False
    self._connection = weakref.ref(connection)

  def _Execute(self, query):
    """Actually executes the query and returns the result of it.

    Arguments:
      @ query: basestring,
        Fully formatted sql statement to execute. In case of unicode, the
        string is encoded to the local character set before it is passed on
        to the server.

    Returns:
      sqlresult.ResultSet instance holding all query result data.
    """
    # TODO(Elmer): Fix this so that arguments can be given independent of the
    # query they belong to. This enables proper SelectTables and enables a host
    # of other escaping things to start working properly.
    #   Refer to MySQLdb.cursor code (~line 151) to see how this works.
    self._LogQuery(query)
    return self.connection.Query(query.strip(), self)

  def _LogQuery(self, query):
    connection = self.connection
    if not isinstance(query, str):
      query = str(query, connection.charset, errors='replace')
    connection.logger.debug(query)
    connection.queries.append(query)

  @staticmethod
  def _StringConditions(conditions, _unused_field_escape):
    if not conditions:
      return '1'
    elif not isinstance(conditions, str):
      return ' AND '.join(conditions)
    return conditions

  @staticmethod
  def _StringFields(fields, field_escape):
    if fields is None:
      return '*'
    elif isinstance(fields, str):
      return field_escape(fields)
    return ', '.join(field_escape(fields, True))

  @staticmethod
  def _StringGroup(group, field_escape):
    if group is None:
      return ''
    elif isinstance(group, str):
      return 'GROUP BY ' + field_escape(group)
    return 'GROUP BY ' + ', '.join(field_escape(group, True))

  @staticmethod
  def _StringLimit(limit, offset):
    if limit is None:
      return ''
    elif offset:
      return 'LIMIT %d OFFSET %d' % (limit, offset)
    return 'LIMIT %d' % limit

  @staticmethod
  def _StringOrder(order, field_escape):
    if order is None:
      return ''
    orders = []
    for rule in order:
      if isinstance(rule, str):
        orders.append(field_escape(rule))
      else:
        orders.append('%s %s' % (field_escape(rule[0]), ('', 'DESC')[rule[1]]))
    return 'ORDER BY ' + ', '.join(orders)

  @staticmethod
  def _StringTable(table, field_escape):
    if isinstance(table, str):
      return field_escape(table)
    return ', '.join(field_escape(table, True))

  def Delete(self, table, conditions, order=None,
             limit=None, offset=0, escape=True):
    """Remove row(s) from table that match conditions, up to limit.

    Arguments:
      table:      string. Name of the table to delete.
      conditions: string/list/tuple (optional).
                  Where statements. Literal as string. AND'd if list/tuple.
                  THESE WILL NOT BE ESCAPED FOR YOU, EVER.
      order:      (nested) list/tuple (optional).
                  Defines sorting of table before updating, elements can be:
                    string: a field to order by (in default database order).
                    list/tuple of two elements:
                      1) string, field name to order by
                      2) bool, revserse; set this to True to reverse the order
      limit:      integer. Defines max number of rows to delete. Default: None.
      offset:     integer (optional). Number of rows to skip, requires limit.
      escape:     boolean. Defines whether table and field names should be
                  escaped. Set this to False if you want to make use of MySQL
                  functions on this query. Default True.

    Returns:
      sqlresult.ResultSet object.
    """
    field_escape = self.connection.EscapeField if escape else lambda x: x
    return self._Execute('delete from %s where %s %s %s' % (
        self._StringTable(table, field_escape),
        self._StringConditions(conditions, field_escape),
        self._StringOrder(order, field_escape),
        self._StringLimit(limit, offset)))

  def Describe(self, table, field=''):
    """Describe table in database or field in table.

    Takes
      table: string. Name of the table to describe.
      field: string (optional). Field name to describe.

    Returns:
      sqlresult.ResultSet object.
    """
    return self._Execute('DESC %s %s' % (
        self._StringTable(table, self.connection.EscapeField),
        self._StringFields(field, self.connection.EscapeField)))

  def Execute(self, query):
    """Executes a raw query."""
    return self._Execute(query)

  def Insert(self, table, values, escape=True):
    """Insert new row into table.

    This method can also perform multi-row insert.
    By default, input strings are quoted, made safe to be inserted into MySQL
    and the Python None-object is translated to MySQL 'NULL'.

    Arguments:
      table:   string. Name of the table to insert into.
      values:  dictionary or list/tuple.
               Dictionary for single inserts:
               * keys:   field names
               * values: field values
               List of dictionaries for a multi-row insert:
               * Each record as a single dictionary.
               * Each dictionary should have the same keys (fields).
      escape:  boolean. Defines whether table names, fields and values should
               be escaped. Set this to False if you want to make use of
               MySQL functions on this query. Default True.

    Returns:
      sqlresult.ResultSet object.
    """
    if not values:
      raise ValueError('Must insert 1 or more value')
    values = self.connection.EscapeValues(values) if escape else values
    table = self.connection.EscapeField(table) if escape else table
    try:
      # Single insert
      values = ', '.join('`%s`=%s' % value for value in values.items())
      query = 'INSERT INTO %s SET %s' % (table, values)
    except AttributeError:
      # Multi-row insert
      fields = ', '.join(map(self.connection.EscapeField, values[0]))
      values = ', '.join('(%s)' % ', '.join(row.itervalues()) for row in values)
      query = 'INSERT INTO %s (%s) VALUES %s' % (table, fields, values)
    return self._Execute(query)

  def Select(self, table, fields=None, conditions=None, order=None,
             group=None, limit=None, offset=0, escape=True, totalcount=False,
             distinct=False):
    """Select fields from table that match the conditions, ordered and limited.

    Arguments:
      table:      string/list/tuple. Table(s) to select fields out of.
      fields:     string/list/tuple (optional). Fields to select. Default '*'.
                  As string, single field name. (autoquoted)
                  As list/tuple, one field name per element. (autoquoted)
                  If the fielname itself is supplied as a tuple,
                  `field` as `name' will be returned where name is the second
                  item in the tuple. (autoquoted)
      conditions: string/list/tuple (optional). SQL 'where' statement.
                  Literal as string. AND'd if list/tuple.
                  THESE WILL NOT BE ESCAPED FOR YOU, EVER.
      order:      (nested) list/tuple (optional).
                  Defines sorting of table before updating, elements can be:
                    string: a field to order by (in default database order).
                    list/tuple of two elements:
                      1) string, field name to order by
                      2) bool, revserse; set this to True to reverse the order
      group:      str (optional). Field name or function to group result by.
      limit:      integer (optional). Defines output size in rows.
      offset:     integer (optional). Number of rows to skip, requires limit.
      escape:     boolean. Defines whether table and field names should be
                  escaped. Set this to False if you want to make use of MySQL
                  functions on this query. Default True.
      totalcount: boolean. If this is set to True, queries with a LIMIT applied
                  will have the full number of matching rows on
                  the affected_rows attribute of the resultset.
      distinct:   bool (optional). Performs a DISTINCT query if set to True.

    Returns:
      sqlresult.ResultSet object.
    """
    field_escape = self.connection.EscapeField if escape else self.NoEscapeField
    result = self._Execute('SELECT %s %s %s FROM %s WHERE %s %s %s %s' % (
        'SQL_CALC_FOUND_ROWS' if totalcount and limit is not None else '',
        'DISTINCT' if distinct else '',
        self._StringFields(fields, field_escape),
        self._StringTable(table, field_escape),
        self._StringConditions(conditions, field_escape),
        self._StringGroup(group, field_escape),
        self._StringOrder(order, field_escape),
        self._StringLimit(limit, offset)))
    if totalcount and limit is not None:
      result.affected = self._Execute('SELECT FOUND_ROWS()')[0][0]
    return result

  def NoEscapeField(self, field, multiple=False):
    """Returns a SQL unescaped field or table name.

    Set multiple = True if field is a tuple of names to be returned.
    If multiple = False, and a tuple is encountered `field` as `name` will be
      returned where the second part of the tuple is the `name` part.
    """
    if not field:
      return ''
    if isinstance(field, str):
      return field
    elif not multiple and isinstance(field, tuple):
      return '%s as %s' % (field[0], field[1])
    return map(self.NoEscapeField, field)

  def SelectTables(self, contains=None, exact=False):
    """Returns table names from the current database.

    Arguments
      % contains: str ~~ ''
        A substring required to be present in all returned table names.
      % exact: bool ~~ False
        Flags whether the string given in contains should be the exact name.

    Returns:
      set: tables names that match the filter.
    """
    if contains:
      contains = self.connection.EscapeValues(contains)
      if exact:
        result = self._Execute('SHOW TABLES LIKE %s' % contains)
      else:
        # Strip quotes from escaped string to allow insertion of wildcards.
        result = self._Execute('SHOW TABLES LIKE "%%%s%%"' % (
            contains.strip("'")))
    else:
      result = self._Execute('SHOW TABLES')
    return set(result[result.fieldnames[0]])

  def Truncate(self, table):
    """Truncate table in database, reducing it to 0 rows.

    Arguments:
      table: string, name of the table to truncate.

    Returns:
      sqlresult.ResultSet object.
    """
    return self._Execute('TRUNCATE %s' % (
        self._StringTable(table, self.connection.EscapeField)))

  def Update(self, table, values, conditions, order=None,
             limit=None, offset=None, escape=True):
    """Updates table records to the new values where conditions are met.

    Arguments:
      table:      string. Name of table to update values in.
      values:     dictionary. Key for fieldname, value for content (autoquoted).
      conditions: string/list/tuple.
                  Where statements. Literal as string. AND'd if list/tuple.
                  THESE WILL NOT BE ESCAPED FOR YOU, EVER.
      order:      (nested) list/tuple (optional).
                  Defines sorting of table before updating, elements can be:
                    string: a field to order by (in default database order).
                    list/tuple of two elements:
                      1) string, field name to order by
                      2) bool, revserse; set this to True to reverse the order
      limit:      integer (optional). Defines max rows to update.
                  Default value for this is None, meaning no limit.
      offset:     integer (optional). Number of rows to skip, requires limit.
      escape:     boolean. Defines whether table names, fields and values should
                  be escaped. Set this to False if you want to make use of
                  MySQL functions on this query. Default True.

    Returns:
      sqlresult.ResultSet object.
    """
    if escape:
      field_escape = self.connection.EscapeField
      values = self.connection.EscapeValues(values)
    else:
      field_escape = lambda x: x

    return self._Execute('UPDATE %s SET %s WHERE %s %s %s' % (
        self._StringTable(table, field_escape),
        ', '.join('`%s`=%s' % value for value in values.items()),
        self._StringConditions(conditions, field_escape),
        self._StringOrder(order, field_escape),
        self._StringLimit(limit, offset)))

  def _ProcessWarnings(self, resultset):
    """Updates messages attribute with warnings given by the MySQL server."""
    db_info = self.connection.Info()
    db_warnings = self.connection.ShowWarnings()
    if db_warnings:
      # This is done in two loops in case Warnings are set to raise exceptions.
      for warning in db_warnings:
        self.connection.logger.warning(
            '%d: %s\nQuery: %s', warning[1], warning[2], resultset.query)
        resultset.warnings.append(warning)
      for warning in db_warnings:
        warnings.warn(warning[-1], self.Warning, 3)

  @property
  def connection(self):
    """Returns the connection that this cursor belongs to."""
    connection = self._connection()
    if connection is None:
      raise self.ProgrammingError('Connection for this cursor closed.')
    return connection

  DatabaseError = pymysql.DatabaseError
  DataError = pymysql.DataError
  Error = pymysql.Error
  IntegrityError = pymysql.IntegrityError
  InterfaceError = pymysql.InterfaceError
  InternalError = pymysql.InternalError
  NotSupportedError = pymysql.NotSupportedError
  OperationalError = pymysql.OperationalError
  ProgrammingError = pymysql.ProgrammingError
  Warning = pymysql.Warning



#!/usr/bin/python3
"""This file contains the connector for Mysql."""
__author__ = 'Jan Klopper (janunderdark.nl)'
__version__ = 0.1

from . import Connector

class Mysql(Connector):
  """Adds MySQL support to connection manager object."""

  def __init__(self, config, options, request, debug=False):
    """Returns a MySQL database connection."""
    self.debug = debug
    self.options = {'host': 'localhost',
                   'user': None,
                   'password': None,
                   'database': ''}
    try:
      from ..libs.sqltalk import mysql
      try:
        self.options = options[self.Name()]
      except KeyError:
        pass
      self.connection = mysql.Connect(
        host=self.options.get('host', 'localhost'),
        user=self.options.get('user'),
        passwd=self.options.get('password'),
        db=self.options.get('database'),
        charset=self.options.get('charset', 'utf8'),
        debug=self.debug)
    except Exception as e:
      raise ConnectionError('Connection to "%s" of type "%s" resulted in: %r' % (self.Name(), type(self), e))

  def Rollback(self):
    with self.connection as cursor:
      return cursor.Execute("ROLLBACK")

  def Disconnect(self):
    """Closes the MySQL connection."""
    if self.debug:
      print('%s closed connection to: %r' % (self.Name(), self.options.get('database')))
    self.connection.close()
    del(self.connection)



#!/usr/bin/python3
"""SQL result abstraction module.

Classes:
  ResultRow: Dict-like object that represents a single database result row.
  ResultSet: Abstraction for a database resultset.

Error Classes:
  Error: Exception base class.
  FieldError: Field- index or name does not exist.
  NotSupportedError: Operation is not supported
"""
__author__ = ('Elmer de Looff <elmer@underdark.nl>',
              'Jan Klopper <jan@underdark.nl>')
__version__ = '1.4'

# Standard modules
import operator

GET_FIELD_NAME = operator.itemgetter(0)


class Error(Exception):
  """Exception base class."""


class FieldError(Error, LookupError):
  """Field- index or name does not exist."""


class NotSupportedError(Error, TypeError):
  """Operation is not supported."""


class ResultRow(object):
  """SQL Result row - an ordered dictionary-like record abstraction.

  ResultRow has two item retrieval interfaces:
    1) Key-based access like that of a dictionary.
    2) Indexed access like a tuple. (field-order is preserved)

  Deleting items from the ResultRow can be done both on index and key. Updating
  or adding fields to the ResultRow can only be done on a key-basis.

  Members:
    % names: tuple (read-only)
      Names for the fields that the ResultRow contains.
  """
  # We expect many ResultRow instances, __slots__ cuts the memory footprint
  # in half for small rows. This seems like a reasonable tradeoff.
  __slots__ = ('_fields', '_values')

  def __init__(self, fields, values):
    """Sets up the ordered dict.

    Arguments:
      @ fields: iterable
        Fieldnames for the SQL result
      @ values: iterable
        Values that belong to the provided fields
    """
    self._fields = list(fields)
    self._values = list(values)

  def __eq__(self, other):
    """Checks equality of the ResultRow to another ResultRow or object.

    A ResultRow can only be equal to another ResultRow, and then only if both
    fieldnames and fieldvalues are the same (data and order).
    """
    return isinstance(other, type(self)) and self.items() == other.items()

  def __getitem__(self, key):
    try:
      if isinstance(key, int):
        return self._values[key]
      else:
        return self._values[self._fields.index(key)]
    except (LookupError, ValueError) as message:
      raise FieldError(message)

  def __repr__(self):
    """Returns a string representation of the ResultRow."""
    return '%s(%s)' % (self.__class__.__name__,
                       ', '.join('%s=%r' % item for item in self.iteritems()))

  def get(self, key, default=None):
    try:
      return self[key]
    except FieldError:
      return default

  @property
  def names(self):
    return self.keys()

  # ############################################################################
  # Iteration on dictionary / record entries
  #
  def __len__(self):
    """Returns the length of the ResultRow."""
    return len(self._values)

  def __iter__(self):
    """Returns an iterator for the values of the ResultRow."""
    return iter(self._values)

  def __reversed__(self):
    """Returns a reversed value iterator."""
    return reversed(self._values)

  def iterkeys(self):
    return iter(self._fields)

  def itervalues(self):
    return iter(self._values)

  def iteritems(self):
    return zip(self._fields, self._values)

  def keys(self):
    return self._fields[:]

  def values(self):
    return self._values[:]

  def items(self):
    return zip(self._fields, self._values)

  # ############################################################################
  # Methods to keep the dictionary neat and ordered
  #
  def __delitem__(self, key):
    """Removes a key or index from the ResultRow."""
    try:
      index = key if isinstance(key, int) else self._fields.index(key)
      del self._fields[index]
      del self._values[index]
    except (LookupError, ValueError):
      raise FieldError('The ResultRow has no field %r' % key)

  def __setitem__(self, field, value):
    """Sets or updates a dictionary value.

    N.B. The implementation for this is slow for large ResultRow objects.
    """
    try:
      self._values[self._fields.index(field)] = value
    except ValueError:
      # The field does not already occur in the ResultRow, add it at the end
      self._fields.append(field)
      self._values.append(value)

  def pop(self, key, *default):
    try:
      index = self._fields.index(key)
      del self._fields[index]
      return self._values.pop(index)
    except ValueError:
      if default:
        return default[0]
      raise FieldError('No field %r in this ResultRow', key)

  def popitem(self):
    """Pops the key,value pair at the end of the dictionary."""
    if not self:
      raise KeyError
    return self._fields.pop(), self._values.pop()


class ResultSet(object):
  """SQL Result set - stores the query, the returned result, and other info.

  ResultSet is created from immutable objects. Once defined, none of its
  attributes can be altered or overwritten. The exception to this is the private
  member _fieldnames which has to be a list for fieldname lookup purposes.

  Members:
    @ affected - int
      Number of rows affected by last action.
    @ charset - str
      Character set used for this connection.
    @ fields - tuple
      Fields in the ResultSet.
    @ insertid - int
      Auto-increment ID that was generated upon the last insert.
    @ query - str
      The executed query that gave this result set.
    @ result:   tuple
      SQL Result set for the last executed query.
    @ _fieldnames - list
      Names of the fields in the result, used for reverse-indexing.
    % fieldnames - tuple (read-only)
      Names of the fields in the result.
  """

  def __init__(self, query='', charset='', result=None, fields=None,
               affected=0, insertid=0, row_class=ResultRow):
    """Initializes a new ResultSet.

    Arguments:
      % affected: int ~~ 0
        Number of affected rows from this operation.
      % charset: str ~~ ''
        Character set used by the connection that executed this operation.
      % fields: tuple of strings ~~ None
        Description of fields involved in this operation.
      % insertid: int ~~ 0
        Auto-increment ID that was generated upon the last insert.
      % query ~~ ''
        The query that was executed for this operation.
      % result ~~ None
        SQL Result set for the this operation.
    """
    self.affected = affected
    self.charset = charset
    self.insertid = insertid
    self.query = query
    self.warnings = []

    if result:
      self.fields = fields
      self.raw = result
      self.result = [row_class(fields, row.values()) for row in result]
    else:
      self.fields = ()
      self.result = []

  def __eq__(self, other):
    """Checks equality of the ResultSet to another ResultSet or object.

    A ResultSet can only be equal to another ResultSet, and then only if all
    their public members compare equal.
    """
    if self is other:
      return True
    elif isinstance(other, self.__class__):
      return (self.affected == other.affected and
              self.insertid == other.insertid and
              self.fields == other.fields and
              self.result == other.result)
    else:
      return False

  def __getitem__(self, item):
    """Returns a row or column from the ResultSet by either index or fieldname.

    Arguments:
      @ item: int / str
        Rownumber or fieldname:
        - If given a rownumber, the corresponding ResultRow is returned.
        - If given a fieldname, a tuple with the field's values is returned.

    Returns:
      ResultRow / tuple: As detailed in the Arguments section.
    """
    try:
      return self.result[item]
    except IndexError:
      raise FieldError('Bad field index: %r.' % item)
    except FieldError:
      raise
    except TypeError:
      # The item type is incorrect, try grabbing a column for this fieldname.
      try:
        index = self._fields.index(item)
        return tuple(row[index] for row in self.result)
      except ValueError:
        raise FieldError('Bad field name: %r.' %  item)

  def __iter__(self):
    """Returns an iterator for the contained ResultRows."""
    return iter(self.result)

  def __len__(self):
    """Returns an integer equal to the number of rows contained."""
    return len(self.result)

  def __nonzero__(self):
    """Boolean truthness of the ResultSet. True if it has 1+ ResultRow"""
    return bool(self.result)

  def __repr__(self):
    """Returns a string representation of the ResultSet."""
    return '%s instance: %d row%s' % (
        self.__class__.__name__, len(self.result), 's'[len(self.result) == 1:])

  def FilterRowsByFields(self, *fields):
    """Yields ResultRows containing only selected fields.

    Arguments:
      @ *fields: list of str
        The fieldnames to filter.

    Raises:
      BadFieldError: One of the given fieldnames did not exist.

    Yields:
      ResultRow: Each ResultRow contains only the filtered fields.
    """
    try:
      indices = tuple(self._fields.index(field) for field in fields)
    except ValueError:
      raise FieldError('Bad fieldnames in filter request.')
    for row in self:
      yield ResultRow(zip(fields, tuple(row[index] for index in indices)))

  def PopField(self, field):
    try:
      self._fields.remove(field)
    except ValueError:
      raise FieldError('Fieldname %r does not occur in the ResultSet.' % field)
    return [row.pop(field) for row in self]

  def PopRow(self, row_index):
    return self.result.pop(row_index)

  @property
  def fieldnames(self):
    """Returns a tuple of the fieldnames that are in this ResultSet."""
    return tuple(self._fields)



#!/usr/bin/python3
"""This module implements the Connection class, which sets up a connection to
a MySQL database. From this connection, cursor objects can be created, which
use the escaping and character encoding facilities offered by the connection.
"""
__author__ = ('Elmer de Looff <elmer@underdark.nl>',
              'Jan Klopper <jan@underdark.nl>')
__version__ = '0.17'

# Standard modules
import pymysql
import logging
import threading
import weakref

# Application specific modules
from pymysql import constants
from pymysql import converters
from . import cursor
from .. import sqlresult


class Connection(pymysql.connections.Connection):
  """MySQL Database Connection Object"""

  def __init__(self, user, passwd, *args, **kwargs):
    """Create a connection to the database. It is strongly recommended
    that you only use keyword parameters. Consult the MySQL C API
    documentation for more information.

    Arguments:
      user:               string, user to connect as.
      passwd:             string, password to use.
      db:                 string, database to use. Default same as user.
      host:               string, host to connect to. Default 'localhost'.
      port:               integer, TCP/IP port to connect to.
      unix_socket:        string, location of unix_socket to use.
      conv:               conversion dictionary, see converters module.
      connect_timeout:    number of seconds to wait before the connection
                          attempt fails.
      compress:           bool, enable compression. Not supported by PyMySQL.
      named_pipe:         if set, a named pipe is used to connect Not supported
                          by PyMySQL.
      init_command:       command which is run once the connection is created
      read_default_file:  file from which default client values are read
      read_default_group: configuration group to use from the default file
      use_unicode:        If True, text-like columns are returned as unicode
                          objects using the connection's character set.
                          Otherwise, text-like columns are returned as strings.
                          Columns are returned as normal strings. Unicode
                          objects will always be encoded to the connection's
                          character set regardless of this setting.
      charset:            If supplied, the connection character set will be
                          changed to this character set (MySQL-4.1 and newer).
                          This enforces use_unicode=True.
      sql_mode:           If supplied, the session SQL mode will be changed to
                          this setting (MySQL-4.1 and newer). For more details
                          and legal values, see the MySQL documentation.
      client_flag:        integer, flags to use or 0.
                          (see MySQL docs or constants/CLIENTS.py)
      ssl:                dictionary or mapping, contains SSL connection
                          parameters; see the MySQL documentation for more
                          details (mysql_ssl_set()).  If this is set, and the
                          client does not support SSL, NotSupportedError will
                          be raised.
      local_infile:       bool, True enables LOAD LOCAL INFILE, False disables.
                          Default False

    There are a number of undocumented, non-standard arguments. See the
    documentation for the MySQL C API for some hints on what they do.
    """
    # Counters, transaction lock & timer
    self.counter_transactions = 0
    self.counter_queries = 0
    self.queries = []
    self.transaction_timer = None
    self.lock = threading.Lock()
    self._charset = None

    # PyMySQL connect args mapping
    kwargs['user'] = user
    kwargs['passwd'] = passwd
    kwargs['host'] = kwargs.get('host', 'localhost')
    kwargs['db'] = kwargs.get('db', user)
    if 'compress' in kwargs:
      raise self.NotSupportedError("PyMySQL doesn't support compression.")
    if 'named_pipe' in kwargs:
      raise self.NotSupportedError("PyMySQL doesn't support named pipes.")
    self.logger = logging.getLogger('mysql_%s' % kwargs['db'])
    if kwargs.pop('debug', False):
      self.debug = True
      self.logger.setLevel(logging.DEBUG)
    else:
      self.debug = False
      self.logger.setLevel(logging.WARNING)
    if kwargs.pop('disable_log', False):
      self.logger.disable_logger = True

    autocommit = kwargs.pop('autocommit', None)
    charset = kwargs.pop('charset', 'utf8')
    sql_mode = kwargs.pop('sql_mode', None)
    # use_unicode = kwargs.pop('use_unicode', False) or bool(charset)

    # The following voodoo is necessary to avoid double references that would
    # prevent a connection object from being finalized and collected properly.
    db = weakref.proxy(self)

    # def _GetUnicodeLiteral():
    #   def UnicodeLiteral(u_string, _dummy=None):
    #     """Returns the SQL (safe) literal for the given unicode object."""
    #     return db.EscapeValues(u_string.encode(db.charset))
    #   return UnicodeLiteral

    def _GetStringDecoder():
      def StringDecoder(string):
        """Returns the unicode codepoints for an encoded bytestream."""
        return string.decode(db.charset)
      return StringDecoder

    encoders = {}
    converts = {}
    conversions = converters.conversions
    self.string_decoder = _GetStringDecoder()

    # if use_unicode:
    #   decoder = self.string_decoder
    #   conversions[constants.FIELD_TYPE.STRING] = decoder
    #   conversions[constants.FIELD_TYPE.VAR_STRING] = decoder
    #   conversions[constants.FIELD_TYPE.VARCHAR] = decoder
    #   conversions[constants.FIELD_TYPE.BLOB] = decoder

    for key, value in conversions.items():
      if not isinstance(key, int):
        encoders[key] = value
      else:
        if isinstance(value, list):
          converts[key] = value[:]
        else:
          converts[key] = value
    kwargs.setdefault('conv', {}).update(converts)

    client_version = tuple(map(int, pymysql.get_client_info().split('.')[:2]))
    kwargs.setdefault('client_flag', 0)
    if client_version >= (4, 1):
      kwargs['client_flag'] |= constants.CLIENT.MULTI_STATEMENTS
    if client_version >= (5, 0):
      kwargs['client_flag'] |= constants.CLIENT.MULTI_RESULTS

    # Done redefining variables for initialization. Engage PyMySQL!
    super(Connection, self).__init__(*args, **kwargs)
    self.encoders = encoders
    # self.encoders[unicode] = self.unicode_literal = _GetUnicodeLiteral()
    self.converter = conversions

    if sql_mode:
      self.SetSqlMode(sql_mode)

    self.charset = charset or self.character_set_name()

    self.transactional = bool(self.server_capabilities &
                              constants.CLIENT.TRANSACTIONS)
    self.autocommit_mode = None
    if autocommit is not None:
      self.autocommit_mode = autocommit
    else:
      self.autocommit_mode = not self.transactional

  def __enter__(self):
    """Refreshes the connection and returns a cursor, starting a transaction."""
    if self.lock.acquire():  # This will block when the lock is in use. In normal situations this should never happen.
      self.counter_transactions += 1
      del self.queries[:]
      self._SetAutocommitState(self.autocommit)
      self.StartTransactionTimer()
      return cursor.Cursor(self)
    raise self.OperationalError(
        'A transaction is already open for this connection.')

  def __exit__(self, exc_type, exc_value, _exc_traceback):
    """End of transaction: commits on success, or rolls back on failure."""
    self.ResetTransactionTimer()
    if exc_type:
      self.rollback()
      if self.debug:
        self.logger.exception(
            'The transaction was rolled back after an exception.\n'
            'Server: %s\nQueries in transaction (last one triggered):\n\n%s',
            self.get_host_info(),
            '\n\n'.join(self.queries))
    else:
      self.commit()
      if self.debug:
        self.logger.debug(
            'Transaction committed (server: %r).', self.get_host_info())
    self.lock.release()

  def CurrentDatabase(self):
    """Return the name of the currently used database"""
    return self.Query('SELECT DATABASE()')[0]

  def EscapeField(self, field, multiple=False):
    """Returns a SQL escaped field or table name.

    Set multiple = True if field is a tuple of names to be escaped.
    If multiple = False, and a tuple is encountered `field` as `name` will be
      returned where the second part of the tuple is the `name` part.
    """
    if not field:
      return ''
    if isinstance(field, str):
      fields = '.'.join('`%s`' % f.replace('`', '``') for f in field.split('.'))
      return fields.replace('`*`', '*')
    elif not multiple and isinstance(field, tuple):
      return '%s as %s' % (self.EscapeField(field[0]), self.EscapeField(field[1]))
    return map(self.EscapeField, field)

  def EscapeValues(self, obj):
    """Escapes any object passed in following the encoders dictionary.

    Sequences and mappings will only have their contents escaped. All strings
    will be encoded to the connection's character set.
    """
    if isinstance(obj, tuple) or isinstance(obj, list):
      return list(map(lambda x: self.escape(x, self.encoders), obj))
    return self.escape(obj, self.encoders)

  def Info(self):
    """Returns a dictionary of MySQL server info and current active database.

    Returns
      dictionary: keys: 'db', 'charset', 'server', 'debug', 'autocommit',
                        'querycount', 'transactioncount'
    """
    return {'db': self.CurrentDatabase(),
            'charset': self._GetCharacterSet(),
            'server': self.ServerInfo(),
            'debug': self.debug,
            'autocommit': self.autocommit_mode,
            'querycount': self.counter_queries,
            'transactioncount': self.counter_transactions}

  def Query(self, query_string, cur=None):
    self.counter_queries += 1
    if isinstance(query_string, str):
      query_string = query_string.encode(self.charset)
    if not cur:
      cur = cursor.Cursor(self)
    cur.execute(query_string)
    stored_result = cur.fetchall()
    if stored_result:
      fields = list(stored_result[0])
    else:
      fields = []
    return sqlresult.ResultSet(
        affected=self.affected_rows(),
        charset=self.charset,
        fields=fields,
        insertid=self.insert_id(),
        query=query_string.decode(self.charset, 'ignore'),
        result=stored_result)

  def ServerInfo(self):
    """Returns a mysql specific set of server information"""
    return self.get_server_info()

  def SetSqlMode(self, sql_mode):
    """Set the connection sql_mode. See MySQL documentation for legal values."""
    self.Query('SET SESSION sql_mode=%s' % self.EscapeValues(sql_mode))

  def ShowWarnings(self):
    """Return detailed information about warnings as a sequence of tuples of
    (Level, Code, Message)."""
    return self.Query('SHOW WARNINGS')

  def StartTransactionTimer(self, delay=60):
    """Writes a warning to the log if the transaction is open too long.

    N.B. The timer is only set when the connection is in debug mode. Calling
    this method on a non-debug connection will do nothing.
    """
    def Warn(caller, delay=delay):
      self.logger.warning('Transaction open for more than %s seconds.', delay)

    if self.debug:
      self.transaction_timer = threading.Timer(delay, Warn)
      self.transaction_timer.daemon = True
      self.transaction_timer.start()

  def ResetTransactionTimer(self):
    """Resets any existing transaction timer."""
    if self.transaction_timer:
      self.transaction_timer.cancel()

  def _GetAutocommitState(self):
    """This returns the current setting for autocommiting transactions."""
    return self.autocommit_mode

  def _GetCharacterSet(self):
    """This configures the character set used by this connection.

    Character sets should be specified as known and supported by MySQL.
    """
    return self._charset

  def _SetAutocommitState(self, state):
    """This sets the autocommit mode on the connection.

    This is False by default if the database supports transactions."""
    try:
      self.ping(reconnect=True)
    except:
      self.connect(sock=None)
    super(Connection, self).autocommit(state)
    self.autocommit_mode = state

  def _SetCharacterSet(self, charset):
    """This sets the character set, refer to _GetCharacterSet for doc."""
    if charset != self._charset:
      super(Connection, self).set_charset(charset)
      self._charset = charset


  # Error classes taken from PyMySQL
  Error = pymysql.Error
  InterfaceError = pymysql.InterfaceError
  DatabaseError = pymysql.DatabaseError
  DataError = pymysql.DataError
  OperationalError = pymysql.OperationalError
  IntegrityError = pymysql.err.IntegrityError
  InternalError = pymysql.InternalError
  ProgrammingError = pymysql.ProgrammingError
  NotSupportedError = pymysql.NotSupportedError
  Warning = pymysql.Warning



#!/usr/bin/python3
"""SQLTalk SQLite interface package."""

# Standard modules
import sqlite3

# Application specific modules
from . import connection

VERSION_INFO = tuple(map(int, sqlite3.version.split('.')))
SQLITE_VERSION_INFO = tuple(map(int, sqlite3.sqlite_version.split('.')))


def Connect(*args, **kwds):
  """Factory function for connection.Connection."""
  kwds['detect_types'] = sqlite3.PARSE_DECLTYPES
  return connection.Connection(*args, **kwds)


def ThreadConnect(*args, **kwds):
  """Factory function for connection.ThreadedConnection."""
  kwds['detect_types'] = sqlite3.PARSE_DECLTYPES
  return connection.ThreadedConnection(*args, **kwds)


DataError = sqlite3.DataError
DatabaseError = sqlite3.DatabaseError
Error = sqlite3.Error
IntegrityError = sqlite3.IntegrityError
InterfaceError = sqlite3.InterfaceError
InternalError = sqlite3.InternalError
NotSupportedError = sqlite3.NotSupportedError
OperationalError = sqlite3.OperationalError



#!/usr/bin/python3
"""uWeb TemplateParser

Classes:
  Parser: Parses a template by replacing tags with their values.

Error classes:
  Error: Base class for all errors generated by this module
  TemplateKeyError: A tagname
  TemplateReadError: Template file could not be read or found.
"""
__author__ = ('Elmer de Looff <elmer@underdark.nl>',
              'Jan Klopper <jan@underdark.nl>')
__version__ = '1.6'

# Standard modules
import os
import re
import urllib.parse as urlparse
from .libs.safestring import *
import hashlib
import itertools
import ast, math

class Error(Exception):
  """Superclass used for inheritance and external exception handling."""


class TemplateKeyError(Error):
  """The replaced tag has no index, key or attribute of the given value."""


class TemplateNameError(Error):
  """The referenced tag does not exist."""


class TemplateFunctionError(Error):
  """The referenced function does not exist."""


class TemplateValueError(Error, ValueError):
  """There is a mismatch on the number of values, or the value is invalid."""


class TemplateTypeError(Error, TypeError):
  """Inappropriate argument type, or wrong number of arguments."""


class TemplateSyntaxError(Error):
  """The template contains illegal syntax."""


class TemplateReadError(Error, IOError):
  """Template file could not be read or found."""


class TemplateEvaluationError(Error):
  """Template condition was not within allowed set of operators."""


class LazyTagValueRetrieval:
  """Provides a means for lazy tag value retrieval.

  This is necessary for instance for TemplateConditional.Expression, where
  lazy retrieval of tag values means that shortcircuit conditional expressions
  become possible.
  """
  def __init__(self, values):
    self._values = values
    self._tags = {}

  # ############################################################################
  # Methods for delayed tag value retrieval
  #
  def __getitem__(self, key):
    return self._tags[key].GetValue(self._values)

  def __setitem__(self, key, value):
    self._tags[key] = value

  # ############################################################################
  # Methods for minimum dictionary likeness
  #
  def __iter__(self):
    return iter(self._tags)

  def iteritems(self):
    """Returns an iterator for the items of the LazyTagValueRetrieval dict."""
    return ((key, self[key]) for key in self)

  def iterkeys(self):
    """Returns an iterator for the keys of the LazyTagValueRetrieval dict."""
    return iter(self)

  def itervalues(self):
    """Returns an iterator for the values of the LazyTagValueRetrieval dict."""
    return (self[key] for key in self)

  def items(self):
    """Returns a list with the items of the LazyTagValueRetrieval dict."""
    return list(self.iteritems())

  def keys(self):
    """Returns a list with the keys of the LazyTagValueRetrieval dict."""
    return list(self)

  def values(self):
    """Returns a list with the values of the LazyTagValueRetrieval dict."""
    return list(self.itervalues())

EVALWHITELIST = {
        'functions': {"abs": abs, "complex": complex, "min": min, "max": max,
                      "pow": pow, "round": round, "len": len, "type": type,
                      "isinstance": isinstance,
                      **{key: value for (key,value) in vars(math).items() if not key.startswith('__')}},
        'operators': (ast.Module, ast.Expr, ast.Load, ast.Expression, ast.Add, ast.And,
                      ast.Sub, ast.UnaryOp, ast.Num, ast.BinOp, ast.Mult, ast.Gt, ast.GtE,
                      ast.Div, ast.Pow, ast.BitOr, ast.BitAnd, ast.BitXor, ast.Lt, ast.LtE,
                      ast.USub, ast.UAdd, ast.FloorDiv, ast.Mod, ast.In, ast.LShift,
                      ast.RShift, ast.Invert, ast.Call, ast.Name, ast.Compare,
                      ast.Eq, ast.NotEq, ast.Not, ast.Or, ast.BoolOp, ast.Str)}

class Parser(dict):
  """A template parser that loads and caches templates and parses them by name.

  After initializing the parser with a search path for templates, new templates
  can be explicitly added by using the `AddTemplate` method, or by using either
  key-based access or using the `Parse` method. These templates are loaded from
  file, though inserted into the Parser cache are Template objects. These are
  constructed into their separate components for faster parsing.

  The `Parse` method takes a template name and any number of keyword arguments.
  The template name is used to fetch the desired Template object from the
  Parser cache (or to load it automatically). This Template object is then
  parsed using the provided keyword arguments.

  Alternatively, there is the `ParseString` method, which works in much the same
  way as the `Parse` method, but the first argument here is a raw template
  string instead.

  Beyond parsing, the parser grants easy access to the TAG_FUNCTIONS dictionary,
  providing the `RegisterFunction` method to add or replace functions in this
  module constant.
  """
  def __init__(self, path='.', templates=(), noparse=False, templateEncoding='utf-8'):
    """Initializes a Parser instance.

    This sets up the template directory and preloads any templates given.

    Arguments:
      % path: str ~~ '.'
        Search path for loading templates using AddTemplate().
      % templates: iter of str ~~ None
        Names of templates to preload.
      % noparse: Bool ~~ False
        Skip parsing the templates to output, instead return their
        structure and replaced values
    """
    super().__init__()
    self.template_dir = path
    self.noparse = noparse
    self.tags = {}
    self.requesttags = {}
    self.astvisitor = AstVisitor(EVALWHITELIST)
    self.templateEncoding = templateEncoding

    for template in templates:
      self.AddTemplate(template)

  def __getitem__(self, template):
    """Retrieves a stored template by name.

    If the template is not already present, it will be loaded from disk.
    The template name will be searched on the defined `template_dir`, if it's
    not found a `TemplateReadError` is raised.

    Arguments:
      @ template: str
        Template name, or the relative path to find it on.

    Raises:
      TemplateReadError: Template name doesn't exist and cannot be loaded.

    Returns:
      Template: A template object, created from a previously loaded file.
    """
    if template not in self:
      self.AddTemplate(template)
    return super().__getitem__(template)

  def AddTemplate(self, location, name=None):
    """Reads the given `template` filename and adds it to the cache.

    The `template` argument should be a path/filename. This will be resolved
    against the configured template directory. The file is parsed and placed in
    the cache using the `template` filename, or the provided `name`.

    Arguments:
      @ location: str
        Location of the template file that should be loaded
      % name: str ~~ None
        Optional name to store the the file as in the cache, instead of the
        template name itself.

    Raises:
      TemplateReadError: When the template file cannot be read
    """
    template_path = os.path.realpath(os.path.join(self.template_dir, location))
    if os.path.commonprefix((template_path, self.template_dir)) != self.template_dir:
      raise TemplateReadError('Could not load template %r' % template_path)
    try:
      self[name or location] = FileTemplate(template_path, parser=self, encoding=None)
    except IOError:
      raise TemplateReadError('Could not load template %r' % template_path)

  def Parse(self, template, **replacements):
    """Returns the referenced template with its tags replaced by **replacements.

    This method automatically loads the referenced template if it doesn't exist.
    The template is loaded from the `template_dir` defined on the instance.

    Arguments:
      @ template: str
        Template name, or the relative path to find it on.
      @ **replacements: dict
        Dictionary of replacement objects. Tags are looked up in here.

    Returns:
      str: The template with relevant tags replaced by the replacement dict.
    """
    if self.tags:
      replacements.update(self.tags)
    if self.requesttags:
      replacements.update(self.requesttags)
    return self[template].Parse(**replacements)

  def ParseString(self, template, **replacements):
    """Returns the given `template` with its tags replaced by **replacements.

    Arguments:
      @ template: str
        The literal template string, where tags are replaced. This is not stored
        in the internal template dictionary, nor is it requested therein.
      @ replacements: dict
        Dictionary of replacement objects. Tags are looked up in here.


    Returns:
      str: template with replaced tags.
    """
    if self.tags:
      replacements.update(self.tags)
    if self.requesttags:
      replacements.update(self.requesttags)
    return Template(template, parser=self).Parse(**replacements)

  @staticmethod
  def RegisterFunction(name, function):
    """Registers a templating `function`, allowing use in templates by `name`.

    Arguments:
      @ name: str
        The name of the template function. This can be used behind a pipe ( | )
      @ function: function
        The function that should be used. Ideally this returns a string.
    """
    TAG_FUNCTIONS[name] = function

  def RegisterTag(self, tag, value, persistent=False):
    """Registers a `value`, allowing use in templates by `tag`.

    Arguments:
      @ tag: str
        The name of the tag
      @ value: str, or function
        Value or function to be executed when replacing this tag
      @ persistent: bool
        will this tag be present for multiple requests?
    """
    storage = self.tags if persistent else self.requesttags
    if ':' not in tag:
      storage[tag] = value
      return
    tag = TemplateTag.FromString('[%s]' % tag)
    # if we are dealing with a tag consisting of multiple path parts, lets reconstruct the path
    obj = storage
    prevnode = tag.name
    d = storage
    for node in tag.indices:
      try:
        node = int(node)
        subtype = SparseList()
      except ValueError:
        subtype = {}

      # add the new sublist to the path if not existant
      if prevnode not in obj:
        obj[prevnode] = subtype

      obj = obj[prevnode]
      prevnode = node
    obj[node] = value

  @classmethod
  def JITTag(cls, function):
    return JITTag(function)

  def ClearRequestTags(self):
    """Resets the non persistent tags to None, is to be called after each
    completed request"""
    self.requesttags = {}

  def SetTemplateEncoding(self, templateEncoding='utf-8'):
    """Allows the user to set the templateEncoding for this parser instance's
    templates. Any template reads, and reloads will be attempted with this
    encoding."""
    self.templateEncoding = templateEncoding

  def SetEvalWhitelist(self, evalwhitelist=None):
    """Allows the user to set the Eval Whitelist which limits the python
    operations allowed within this templateParsers Context. These are usually
    triggered by If/Elif conditions and the like."""
    self.astvisitor = AstVisitor(evalwhitelist)

  TemplateReadError = TemplateReadError


class Template(list):
  """Contained for template parts, allowing for rich content construction."""
  FUNCTION = re.compile(r'\{\{\s*(.*?)\s*\}\}')
  # For a full tag syntax explanation, refer to the TAG regex in TemplateTag.
  TAG = re.compile("""
      (\[\w+                      # Tag start and alphanum tagname
        (?:(?::[\w\-\.]+)+)?           # 0+ indices, alphanum with dashes
        (?:(?:\|[\w-]+              # 0+ functions, alphanum with dashes
          (?:\([^()]*?\))?            # closure parentheses and arguments
        )+)?                        # end of function block
      \])                         # end of tag""",
      re.VERBOSE)

  def __init__(self, raw_template, parser=None):
    """Initializes a Template from a string.

    Arguments:
      @ raw_template: str
        The string that represents the template.
      % parser: Parser ~~ None
        An optional parser instance that is necessary to enable support for
        adding files to the current template. This is used by {{ inline }}.
    """
    super().__init__()
    self.parser = parser
    self.scopes = [self]
    self.AddString(raw_template)

  def __eq__(self, other):
    """Returns the equality to another Template.

    Two templates are equal if they are of the same type, and have the same
    content for their unparsed template; or string representation.
    """
    return isinstance(other, Template) and str(other) == str(self)

  def __mod__(self, kwds):
    """Syntactic sugar that enables percent-sign template parsing.

    The provided keywords MUST be in a dictionary.
    """
    return self.Parse(**kwds)

  def __repr__(self):
    return '%s(%s)' % (type(self).__name__, list(self))

  def __str__(self):
    return ''.join(map(str, self))

  def AddFile(self, name):
    """Extends the Template by reading template data from a file.

    The file is loaded through the Parser instance associated with the template.
    If there is none associated, this will raise a TypeError.

    Raises:
      TemplateReadError: The template file could not be read by the Parser.
      TypeError: There is no parser associated with the template.
    """
    if self.parser is None:
      raise TypeError('The template requires parser for adding template files.')
    return self._AddToOpenScope(self.parser[name])

  def AddString(self, raw_template, filename=None):
    """Extends the Template by adding a raw template string.

    The given template is parsed and added to the existing template.

    Raises:
      TemplateSyntaxError: Unbalanced number of scopes in added template.
    """
    scope_depth = len(self.scopes)
    nodes = self.FUNCTION.split(raw_template)
    for index, node in enumerate(nodes):
      if index % 2:
        self._ExtendFunction(node)
      else:
        self._ExtendText(node)
    if len(self.scopes) != scope_depth:
      scope_diff = len(self.scopes) - scope_depth
      if scope_diff < 0:
        raise TemplateSyntaxError('Closed %d scopes too many in "%s"' % (abs(scope_diff), filename or raw_template))
      raise TemplateSyntaxError('TemplateString left %d open scopes in "%s"' % (scope_diff, filename or raw_template))

  def Parse(self, returnRawTemplate=False, **kwds):
    """Returns the parsed template as HTMLsafestring.

    The template is parsed by parsing each of its members and combining that.
    """
    htmlsafe = HTMLsafestring('').join(HTMLsafestring(tag.Parse(**kwds)) for tag in self)
    htmlsafe.content_hash = hashlib.md5(htmlsafe.encode()).hexdigest()
    if returnRawTemplate:
      raw = HTMLsafestring(self)
      raw.content_hash = htmlsafe.content_hash
      return raw

    if self.parser and self.parser.noparse:
      # Hash the page so that we can compare on the frontend if the html has changed
      htmlsafe.page_hash = hashlib.md5(HTMLsafestring(self).encode()).hexdigest()
      # Hashes the page and the content so we can know if we need to refresh the page on the frontend
      htmlsafe.tags = {}
      for tag in self:
        if isinstance(tag, TemplateConditional):
          for flattend_branch in list(itertools.chain(*tag.branches)):
            for branch_tag in flattend_branch:
              if isinstance(branch_tag, TemplateTag):
                htmlsafe.tags[str(branch_tag)] = branch_tag.Parse(**kwds)
        if isinstance(tag, TemplateTag):
          htmlsafe.tags[str(tag)] = tag.Parse(**kwds)
    return htmlsafe

  @classmethod
  def TagSplit(cls, template):
    """Yields the TemplateTag and TemplateText nodes from a template string."""
    for index, node in enumerate(cls.TAG.split(template)):
      if index % 2:
        yield TemplateTag.FromString(node)
      elif node:
        yield TemplateText(node)

  def _ExtendFunction(self, nodes):
    """Processes a function node and adds its results to the Template.

    For loops, a new scope level is opened by adding the TemplateLoop to the
    `scopes` instance attribute. Upon finding the end of a loop, the topmost
    scope is removed, provided it is a TemplateLoop scope. If it is not,
    TemplateSyntaxError is raised.

    Raises:
      TemplateSyntaxError: Unexpected / unknown command or otherwise bad syntax.
    """
    nodes = nodes.split()
    function = nodes.pop(0)
    try:
      getattr(self, '_TemplateConstruct%s' % function.title())(*nodes)
    except AttributeError:
      raise TemplateSyntaxError('Unknown template function {{ %s }}%s' %
        (function,
         ' in template "%s"' % self._template_path if self._template_path else ''))

  def _ExtendText(self, node):
    """Processes a text node and adds its tags and texts to the Template."""
    for node in self.TagSplit(node):
      self._AddToOpenScope(node)

  # ############################################################################
  # Template syntax constructs
  #

  def _TemplateConstructInline(self, name):
    """Processing for {{ inline }} template syntax."""
    self.AddFile(name)

  def _TemplateConstructFor(self, *nodes):
    """Processing for {{ for }} template syntax."""
    self._StartScope(TemplateLoop(nodes[-1], nodes[:-2]))

  def _TemplateConstructEndfor(self):
    """Processing for {{ endfor }} template syntax."""
    self._CloseScope(TemplateLoop)

  def _TemplateConstructIf(self, *nodes):
    """Processing for {{ if }} template syntax."""
    self._StartScope(TemplateConditional(' '.join(nodes),
        self.parser.astvisitor if self.parser else AstVisitor(EVALWHITELIST)))

  def _TemplateConstructIfpresent(self, *nodes):
    """Processing for {{ ifpresent }} template syntax."""
    self._StartScope(TemplateConditionalPresence(' '.join(nodes),
        self.parser.astvisitor if self.parser else AstVisitor(EVALWHITELIST)))

  def _TemplateConstructIfnotpresent(self, *nodes):
    """Processing for {{ ifnotpresent }} template syntax."""
    self._StartScope(TemplateConditionalNotPresence(' '.join(nodes),
        self.parser.astvisitor if self.parser else AstVisitor(EVALWHITELIST)))

  def _TemplateConstructElif(self, *nodes):
    """Processing for {{ elif }} template syntax."""
    self._VerifyOpenScope(TemplateConditional)
    self.scopes[-1].Elif(' '.join(nodes))

  def _TemplateConstructElse(self):
    """Processing for {{ else }} template syntax."""
    self._VerifyOpenScope(TemplateConditional)
    self.scopes[-1].Else()

  def _TemplateConstructEndif(self):
    """Processing for {{ endif }} template syntax."""
    self._CloseScope(TemplateConditional)

  # ############################################################################
  # Methods for scope management
  #
  def _AddToOpenScope(self, item):
    """Adds a template part to the current open scope."""
    self.scopes[-1].append(item)

  def _CloseScope(self, scope_cls):
    """Closes the current open scope, if it's of the given scope type.

    If the open scope is not an instance of the given scope class,
    TemplateSyntaxError is raised.
    """
    if not isinstance(self.scopes[-1], scope_cls):
      raise TemplateSyntaxError('Tried to close %s, but open scope is %s' % (
          scope_cls.__name__, type(self.scopes[-1]).__name__))
    self.scopes.pop()

  def _StartScope(self, scope):
    """Adds the given part to the template and adds it as new current scope."""
    self._AddToOpenScope(scope)
    self.scopes.append(scope)

  def _VerifyOpenScope(self, scope_cls):
    """Verifies the given `scope_cls` is the current open scope.

    If this is not the case, TemplateSyntaxError is raised.
    """
    if not isinstance(self.scopes[-1], scope_cls):
      raise TemplateSyntaxError('Expected open scope %s, but scope is %s' % (
          scope_cls.__name__, type(self.scopes[-1]).__name__))


class FileTemplate(Template):
  """Template class that loads from file."""
  def __init__(self, template_path, parser=None, encoding='utf-8'):
    """Initializes a FileTemplate based on a given template path.

    Arguments:
      @ template_path: str
        A string to begin a template with. This is parsed and used to build the
        initial raw template from.
      % parser: Parser ~~ None
        An optional parser instance that is necessary to enable support for
        adding files to the current template. This is used by {{ inline }}.
    """
    self._template_path = template_path
    self.parser = parser
    self.templateEncoding = encoding or (self.parser.templateEncoding if self.parser else 'utf-8')
    try:
      self._file_name = os.path.abspath(template_path)
      self._file_mtime = os.path.getmtime(self._file_name)
      # self.parser can be None in which case we default to utf-8
      with open(self._file_name, encoding=self.templateEncoding) as templatefile:
        raw_template = templatefile.read()
      super().__init__(raw_template, parser=parser)
    except (IOError, OSError):
      raise TemplateReadError('Cannot open: %r' % template_path)

  def Parse(self, **kwds):
    """Returns the parsed template as SafeString.

    The template is parsed by parsing each of its members and combining that.
    """
    self.ReloadIfModified()
    try:
      result = super().Parse(**kwds)
    except TemplateFunctionError as error:
      raise TemplateFunctionError('%s in %s' % (error, self._template_path))
    if self.parser and self.parser.noparse:
      return {'template': self._templatepath[len(self.parser.template_dir):],
              'replacements': result.tags,
              'content_hash':result.content_hash,
              'page_hash': result.page_hash}
    return result

  def ReloadIfModified(self):
    """Reloads the template file if it was modified on disk.

    If the template is not present, cannot be read, or there is another error
    accessing the file, the operation is aborted and the old template is left
    in place.

    If the new template has a syntax error or other problem during loading,
    that error *will* be raised.
    """
    try:
      mtime = os.path.getmtime(self._file_name)
      if mtime > self._file_mtime:
        with open(self._file_name, encoding=self.templateEncoding) as templatefile:
          template = templatefile.read()
        del self[:]
        self.scopes = [self]
        self.AddString(template, self._file_name)
        self._file_mtime = mtime
    except (IOError, OSError):
      # File cannot be stat'd or read. No longer exists or we lack permissions.
      # We shouldn't error in this case, but carry on with the template we have.
      pass


class TemplateConditional(object):
  """A template construct to control flow based on the value of a tag."""
  def __init__(self, expr, astvisitor):
    self.branches = []
    self.default = None
    self.NewBranch(expr)
    self.astvisitor = astvisitor

  def __repr__(self):
    repr_branches = []
    for expr, branch in self.branches:
      clause = 'IF' if not repr_branches else 'ELIF'
      repr_branches.append('%s %r { %r }' % (clause, expr, branch))
    if self.default:
      repr_branches.append(' ELSE { %r }' % self.default)
    return '%s(%s)' % (type(self).__name__, ''.join(repr_branches))

  def __str__(self):
    repr_branches = []
    for expr, branch in self.branches:
      clause = 'if' if not repr_branches else 'elif'
      repr_branches.append('{{ %s %s }}%s' % (
          clause, ''.join(map(str, expr)), ''.join(map(str, branch))))
    if self.default:
      repr_branches.append('{{ else }}%s' % ''.join(map(str, self.default)))
    repr_branches.append('{{ endif }}')
    return '\n' + '\n'.join(repr_branches)

  def append(self, part):
    """Appends a template part to the current open conditional clause.

    Conditional clauses are not explicitly closed, a new one is simply stacked
    on top. Whenever the {{ else }} statement is found, all append actions will
    append to the else clause.
    """
    if self.default is not None:
      self.default.append(part)
    else:
      self.branches[-1][1].append(part)

  def Elif(self, expr):
    """Starts an `elif` clause.

    This raises TemplateSyntaxError if the `else` clause is already started.
    """
    if self.default is not None:
      raise TemplateSyntaxError('{{ elif }} clause may not follow {{ else }}.')
    self.NewBranch(expr)

  def Else(self):
    """Starts the `else` clause.

    This raises TemplateSyntaxError if the `else` clause is already started.
    """
    if self.default is not None:
      raise TemplateSyntaxError('Only one {{ else }} clause is allowed.')
    self.default = []

  def Expression(self, expr, **kwds):
    """Returns the evaluated result of a tag expression."""
    nodes = []
    local_vars = LazyTagValueRetrieval(kwds)
    for num, node in enumerate(expr):
      if isinstance(node, TemplateTag):
        node_name = '__tmpl_var_%d' % num
        local_vars[node_name] = node
        nodes.append(node_name)
      else:
        nodes.append(node)
    try:
      return LimitedEval(''.join(nodes), self.astvisitor, local_vars)
    except NameError as error:
      raise TemplateNameError(str(error).capitalize() + '. Try it as [tagname]?')
    except SyntaxError as error:
      raise TemplateSyntaxError('%s while evaluating: %s' % (str(error).capitalize(), ''.join(nodes)))

  def NewBranch(self, expr):
    """Begins a new branch based on the given expression."""
    self.branches.append((tuple(Template.TagSplit(expr)), []))

  def Parse(self, **kwds):
    """Returns the TemplateConditional parsed as string.

    One by one, the `if` clause and optional `elif` clauses are evaluated.
    Their strings are parsed (template tags replaced, though functions are NOT
    processed) and the resulting expression evaluated. Strings passed into the
    templateparser will be strings for evaluation (not literal code), so this
    is safe with regards to users executing code in the templateparser scope.

    Whenever a boolean True value is returned from eval, the corresponding
    branch is parsed and returned. When none of the `if` or `elif` clauses
    is True, the `else` branch is parsed and returned (where available, if no
    `else` branch exists '' is returned.
    """
    for expr, branch in self.branches:
      if self.Expression(expr, **kwds):
        return ''.join(part.Parse(**kwds) for part in branch)
    if self.default:
      return ''.join(part.Parse(**kwds) for part in self.default)
    return ''


class TemplateConditionalPresence(TemplateConditional):
  """A template construct to safely check for the presence of tags."""

  @staticmethod
  def Expression(tags, **kwds):
    """Checks the presence of all tags named on the branch."""
    try:
      for tag in tags:
        tag.GetValue(kwds)
      return True
    except (TemplateKeyError, TemplateNameError):
      return False

  def NewBranch(self, tags):
    """Begins a new branch based on the given tags."""
    self.branches.append((list(map(TemplateTag.FromString, tags.split())), []))


class TemplateConditionalNotPresence(TemplateConditionalPresence):
  """A template construct to safely check for the presence of tags."""

  @staticmethod
  def Expression(tags, **kwds):
    """Checks the presence of all tags named on the branch."""
    try:
      for tag in tags:
        tag.GetValue(kwds)
      return False
    except (TemplateKeyError, TemplateNameError):
      return True


class TemplateLoop(list):
  """Template loops are used to repeat a portion of template multiple times.

  Upon parsing, the loop tag is retrieved, and for each of its members, the
  items in the loop are parsed. The loop variable is made available as the given
  alias, which itself can be referenced as a tag in the loop body.
  """
  def __init__(self, tag, aliases):
    """Initializes a TemplateLoop instance.

    Arguments:
      @ tag: str
        The tag to retrieve the iterable from.
      @ aliases: *str
        The alias(es) under which the loop variable should be made available.
    """
    try:
      tag = TemplateTag.FromString(tag)
    except TemplateSyntaxError:
      raise TemplateSyntaxError('Tag %r in {{ for }} loop is not valid' % tag)

    super().__init__()
    self.aliases = ''.join(aliases).split(',')
    self.aliascount = len(self.aliases)
    self.tag = tag

  def __repr__(self):
    return '%s(%s)' % (type(self).__name__, list(self))

  def __str__(self):
    return '\n{{ for %s in %s }}%s\n{{ endfor }}' % (
        ', '.join(self.aliases), self.tag, ''.join(map(str, self)))

  def Parse(self, **kwds):
    """Returns the TemplateLoop parsed as string.

    Firstly, the value for the loop tag is retrieved. For each item in this
    iterable, all members of the TemplateLoop body will be parsed, with the
    item from the iterable added to the replacements dict as alias(es).
    """
    output = []
    replacements = kwds.copy()
    for item in self.tag.Iterator(**kwds):
      if self.aliascount == 1:
        replacements[self.aliases[0]] = item
      else:
        try:
          if self.aliascount != len(item):
            raise TemplateValueError('Cannot unpack %d values into %d tags' % (
                len(item), self.aliascount))
        except TypeError:
          raise TemplateValueError(
              'Cannot unpack %s into %d tags' % (type(item), self.aliascount))
        replacements.update(zip(self.aliases, item))
      output.append(''.join(tag.Parse(**replacements) for tag in self))
    return ''.join(output)


class TemplateTag(object):
  """Template tags are used for dynamic placeholders in templates.

  Their final value is determined during parsing. For more explanation on this,
  refer to the documentation for Parse().
  """
  PFX_INDEX = ':'
  PFX_FUNCT = '|'
  TAG = re.compile("""
      \[                  # Tag starts with opening bracket
        (\w+)               # Capture tagname (1+ alphanum length)
        ((?::[\w\-\.]+)+)?     # Capture 0+ indices (1+ alphanum+dashes length)
        ((?:\|[\w-]+        # Capture 0+ functions (1+ alphanum+dashes length)
          (?:\([^()]*?\))?    # Functions may be closures with arguments.
        )+)?                # // end of optional functions
      \]                  # // end of tag""",
      re.VERBOSE)
  FUNC_FINDER = re.compile('\|([\w-]+(?:\([^()]*?\))?)')
  FUNC_CLOSURE = re.compile('(\w+)\((.*)\)')
  ALLOWPRIVATE = False # will we allow access to private members for object lookup

  def __init__(self, name, indices=(), functions=()):
    """Initializes a TemplateTag instant.

    Arguments:
      @ name: str
        The name of the tag, to retrieve it from the replacements dictionary.
      % indices: iterable ~~ None
        Indices that should be applied to arrive at the proper tag value.
      % functions: iterable ~~ None
        Names of template functions that should be applied to the value.
    """
    self.name = name
    self.indices = (indices if self.ALLOWPRIVATE else [
        index for index in indices
        if not index.startswith('_') or not index.endswith('_')
    ])
    self.functions = functions

  def __repr__(self):
    return '%s(%r)' % (type(self).__name__, str(self))

  def __str__(self):
    return '[%s%s%s]' % (
        self.name,
        ''.join(self.PFX_INDEX + index for index in self.indices),
        ''.join(self.PFX_FUNCT + func for func in self.functions))

  @classmethod
  def _GetFunctions(cls, raw_functions):
    if raw_functions:
      return tuple(cls.FUNC_FINDER.findall(raw_functions))
    return ()

  @classmethod
  def _GetIndices(cls, raw_indices):
    if raw_indices:
      return raw_indices.lstrip(cls.PFX_INDEX).split(cls.PFX_INDEX)
    return ()

  @classmethod
  def FromString(cls, tag):
    """Returns a TemplateTag object which is parsed from the given string.

    A tag's formatting restrictions are as follows:
      * The whole tag is delimited by square brackets: []
      * Indices are separated by colons, :, multiple are allowed
      * Functions are prefixed by pipes, |, multiple are allowed
      * In addition to the characters stated above, tags may contain only
        alphanumeric values, underscores and dashes. Spaces are _not_ allowed.
    """
    try:
      name, indices, functions = cls.TAG.match(tag).groups()
      return cls(name, cls._GetIndices(indices), cls._GetFunctions(functions))
    except AttributeError:
      raise TemplateSyntaxError('Invalid Tag syntax: %r' % tag)

  def GetValue(self, replacements):
    """Returns the value for the tag, after reducing indices.

    For a tag with indices, these are looked up one after the other, each index
    being that of the next step. [tag:0:0] for with a keyword tag=[['foo']]
    would given 'foo' as the value for the tag.
    """
    try:
      value = replacements[self.name]
      for index in self.indices:
        value = self._GetIndex(value, index)
      if isinstance(value, JITTag):
        return value()
      return value
    except KeyError:
      raise TemplateNameError('No replacement with name %r' % self.name)
    except TemplateKeyError as error:
      raise TemplateKeyError('%s on %r' % (error, self.name))

  @classmethod
  def ApplyFunction(cls, func, value):
    closure = cls.FUNC_CLOSURE.match(func)
    try:
      if not closure:
        return TAG_FUNCTIONS[func](value)
      func, args = closure.groups()
      #XXX(Elmer): This uses eval, it's so much easier than lexing and parsing
      # the regex leading up to this point make sure no function calls end up in
      # here, nor variables, Math might show up though
      args = eval(args + ',', {'__builtins__': {}}, {}) if args.strip() else ()
      return TAG_FUNCTIONS[func](*args)(value)
    except SyntaxError:
      raise TemplateSyntaxError('Invalid argument syntax: %r' % args)
    except TypeError as err_obj:
      raise TemplateTypeError(
          ('Templatefunction raised an TypeError %s(%s) ' % (func, value), err_obj))
    except KeyError as err_obj:
      raise TemplateFunctionError(
          'Unknown template tag function %r' % err_obj.args[0])
    except NameError as err_obj:
      raise TemplateNameError(
          'Access to scope outside of parser variables is not allowed: %r' % err_obj.args[0])

  def Parse(self, **kwds):
    """Returns the parsed string of the tag, using given replacements.

    Firstly, the tag's name is retrieved from the given replacements dictionary.

    After that, for a tag with indices, these are looked up one after the other,
    each index being that of the next step. `[tag:0:0]` with a replacements
    dictionary {tag=[['foo']]} will reduce to 'foo' as the value for the tag.

    After reducing indexes, functions are applied. As before, functions are
    applied one after the other. The second function works on the result of the
    first. If no functions are defined for a tag, the default tag function
    will be applied. SafeString objects are exempt from this default function;
    They will only be acted upon by functions as specified in the tag.

    All tag functions are derived from the module constant TAG_FUNCTIONS, and
    are looked up when requested. This means that if a function is changed after
    the template has been created, the new function will be used instead.
    """
    try:
      value = self.GetValue(kwds)
    except (TemplateKeyError, TemplateNameError):
      # On any failure to get the given index, return the unmodified tag.
      return str(self)
    # Process functions, or apply default if value is not Basesafestring
    if self.functions:
      for func in self.functions:
        try:
          value = self.ApplyFunction(func, value)
        except TemplateFunctionError as error:
          raise TemplateFunctionError('%s on %s' % (error, self))
        except TemplateSyntaxError as error:
          raise TemplateSyntaxError('%s on %s' % (error, self))
    if not isinstance(value, Basesafestring):
      value = TAG_FUNCTIONS['default'](value)
    return value

  def Iterator(self, **kwds):
    """Parses the tag for iteration purposes.

    Functions are processed, but no defaults or other conversion is done. Tags
    that cannot be resolved result in empty iterators.
    """
    try:
      value = self.GetValue(kwds)
    except TemplateKeyError:
      # On any failure to get the given index, return an empty iterator
      return ()
    for func in self.functions:
      value = TAG_FUNCTIONS[func](value)
    return iter(value)

  @staticmethod
  def _GetIndex(haystack, needle):
    """Returns the `needle` from the `haystack` by index, key or attribute name.

    Arguments:
      @ haystack: obj
        The searched object; iterable, mapping or any kind of object.
      @ needle: str
        The index, key or attribute name to find on the haystack.

    Raises:
      TemplateKeyError: One or more of the given needles don't exist.

    Returns:
      obj: the object existing on `needle` in `haystack`.
    """
    try:
      if needle.isdigit():
        try:
          # `needle` is a number; likely an index or a numeric dict-key.
          return haystack[int(needle)]
        except KeyError:
          # `haystack` should be a dict; numeric attributes are invalid syntax.
          return haystack[needle]
      try:
        # `needle` is a string; either a dict-key, or an attribute name.
        return haystack[needle]
      except (KeyError, TypeError):
        # KeyError, `haystack` has no key `needle` but may have matching attr.
        # TypeError: `haystack` is no mapping but may have a matching attr.
        return getattr(haystack, needle)
    except (AttributeError, LookupError):
      raise TemplateKeyError('Item has no index, key or attribute %r' % needle)


class TemplateText(str):
  """A raw piece of template text, upon which no replacements will be done."""
  def __new__(cls, string):
    return super().__new__(cls, string)

  def __repr__(self):
    """Returns the object representation of the TemplateText."""
    return '%s(%r)' % (type(self).__name__, str(self))

  def Parse(self, **_kwds):
    """Returns the string value of the TemplateText."""
    return str(self)


class JITTag(object):
  """This is a template Tag which is only evaulated on replacement.
  It is usefull for situations where not all all of this functions input vars
  are available just yet.
  """

  def __init__(self, function):
    """Stores the function for later use"""
    self.wrapped = function
    self.result = None # cache for results
    self.called = False # keep score of result cache usage, None and False might be correct results in the cache

  def __call__(self):
    """Returns the output of the earlier wrapped function"""
    if not self.called:
      self.result = self.wrapped()
    self.called = True
    return self.result


class SparseList(list):
  """A spare list implementation to allow us to set the nth item on a list"""
  def __setitem__(self, index, value):
    missing = index - len(self) + 1
    if missing > 0:
      self.extend([None] * missing)
    super().__setitem__(index, value)

  def __getitem__(self, index):
    """Return the value at the index, and None of that index was not available
    instead of raisig IndexError
    """
    try:
      return super().__getitem__(index)
    except IndexError:
      return None


class AstVisitor(ast.NodeVisitor):
  def __init__(self, whitelists):
    self.whitelists = whitelists

  def visit(self, node):
    if not isinstance(node, self.whitelists['operators']):
      raise TemplateEvaluationError('`%s` is not an allowed operation' % node)
    return super().visit(node)

  def visit_Call(self, call):
    """Filter calls"""
    if call.func.id not in self.whitelists['functions']:
      raise TemplateEvaluationError('`%s` is not an allowed function call' % call.func.id)

def LimitedEval(expr, astvisitor, evallocals = {}):
  tree = ast.parse(expr, mode='eval')
  astvisitor.visit(tree)
  return eval(compile(tree, "<string>", "eval"),
      astvisitor.whitelists['functions'],
      evallocals)


TAG_FUNCTIONS = {
    'default': lambda d: HTMLsafestring('') + d,
    'html': lambda d: HTMLsafestring('') + d,
    'htmlsource': lambda d: HTMLsafestring(d, unsafe=True),
    'raw': lambda d: Unsafestring(d),
    'url': lambda d: HTMLsafestring(URLqueryargumentsafestring(d, unsafe=True)),
    'type': type,
    'items': lambda d: list(d.items()),
    'values': lambda d: list(d.values()),
    'sorted': sorted,
    'len': len}



#!/usr/bin/python
"""SQLTalk MySQL interface package.

Functions:
  Connect: Connects to a MySQL server and returns a connection object.
           Refer to the documentation enclosed in the connections module for
           argument information.
"""
__author__ = 'Jan Klopper <jan@underdark.nl>'
__version__ = '0.10'

# Application specific modules
from . import connection

def Connect(*args, **kwargs):
  """Factory function for connection.Connection."""
  return connection.Connection(*args, **kwargs)



#!/usr/bin/python3
"""This file contains all the connectionManager classes that interact with
databases, restfull apis, secure cookies, config files etc."""
__author__ = 'Jan Klopper (janunderdark.nl)'
__version__ = 0.2

import sys

import uweb3

from .connectors import *

class ConnectionError(Exception):
  """Error class thrown when the underlying connectors thrown an error on
  connecting."""

class ConnectionManager(object):
  """This is the connection manager object that is handled by all Model Objects.
  It finds out which connection was requested by looking at the call stack, and
  figuring out what database type the model class calling it belongs to.

  Connected databases are stored and reused.
  On delete, the databases are closed and any lingering transactions are
  committed. to complete the database writes.
  """

  DEFAULTCONNECTIONMANAGER = None

  def __init__(self, config, options, debug, requestdepth=3, requestmaxdept=100):
    """Initializes the ConnectionManager

     Arguments:
      % config: reference to config parser
      config instance
      % options: reference to config parsers dict
      options for the various settings provided in the config
      % debug: bool, Optional defaults to False
      Outputs extra debugging information if set to True
      % requestdept: int, indicates how many stack layers we should start at
      lookin up the stack to find our request object. Optional defaults to 2.
      % requestdept: int, requestmaxdepth indicates how many stack layers at
      maximum we should start at lookin up the stack to find our request object.
      Optional defaults to 100.

    """
    self.__connectors = {} # classes
    self.__connections = {} # instances
    self.config = config
    self.options = options
    self.debug = debug
    self.LoadDefaultConnectors()
    self.requestdepth = requestdepth
    self.requestmaxdepth = requestmaxdept

  def LoadDefaultConnectors(self):
    """Populates the list of Connectors with the default available connectors"""
    self.RegisterConnector(SignedCookie)
    self.RegisterConnector(Mysql, True)
    self.RegisterConnector(Sqlite)
    self.RegisterConnector(Mongo)
    self.RegisterConnector(SqlAlchemy)

  def RegisterConnector(self, handler, default=False):
    """Make the ConnectonManager aware of a new type of connector.

    Arguments:
      % handler: class
      Reference to the class that will handle the connections
      % default: bool, Optional defaults to False
      Should this Connector be considers the default connector?
    """
    if default:
      self.DEFAULTCONNECTIONMANAGER = handler.Name()
    self.__connectors[handler.Name()] = handler

  def RelevantConnection(self, level=2):
    """Returns the relevant database connection dependant on the caller model
    class.

    If the caller model cannot be determined, the 'relational' database
    connection is returned as a fallback method.

    Level indicates how many stack layers we should go up to find the current
    caller_class which indicates our connector type. Defaults to 2.

    When no connection can be found or made due to a missing request from this
    context a TypeError will be raised.

    When no connection can be found or made Due to a missing connector class a
    TypeError will be raised.
    """
    # Figure out caller type or instance
    # pylint: disable=W0212
    #TODO use inspect module instead, and iterate over frames
    caller_locals = sys._getframe(level).f_locals
    # pylint: enable=W0212
    # Caller might be a Class or Class instance
    if 'self' in caller_locals:
      caller_cls = type(caller_locals['self'])
    else:
      caller_cls = caller_locals.get('cls', type)
    # Decide the type of connection to return for this caller
    con_type = (caller_cls._CONNECTOR if hasattr(caller_cls, '_CONNECTOR') else
                self.DEFAULTCONNECTIONMANAGER)
    if (con_type in self.__connections and
        hasattr(self.__connections[con_type], 'connection')):
      return self.__connections[con_type].connection

    try:
      # instantiate a connection
      self.__connections[con_type] = self.__connectors[con_type](
          self.config, self.options, self.request, self.debug)
      return self.__connections[con_type].connection
    except KeyError as error:
      raise TypeError('No connector for: %r, available: %r, %r' % (
          con_type, self.__connectors, error))

  @property
  def request(self):
    """Returns the request object as looked up in the stack.

    When no connection can be found or made due to a missing request from this
    context a TypeError will be raised.

    When no connection can be found or made Due to a missing connector class a
    TypeError will be raised.
    """
    requestdepth = self.requestdepth
    while requestdepth < self.requestmaxdepth:
      try:
        parent = sys._getframe(requestdepth).f_locals['self']
        if isinstance(parent, uweb3.PageMaker) and hasattr(parent, 'req'):
          request = parent.req
          if self.debug:
            print('request object found at stack level %d' % requestdepth)
          return request
      except (KeyError, AttributeError, ValueError):
        pass
      requestdepth = requestdepth + 1
    raise TypeError('No request could be found in call Stack.')

  def __enter__(self):
    """Proxies the transaction to the underlying relevant connection."""
    return self.RelevantConnection().__enter__()

  def __exit__(self, *args):
    """Proxies the transaction to the underlying relevant connection."""
    return self.RelevantConnection().__exit__(*args)

  def __getattr__(self, attribute):
    return getattr(self.RelevantConnection(), attribute)

  def RollbackAll(self):
    """Performs a rollback on all connectors with pending commits."""
    if self.debug:
      print('Rolling back uncommited transaction on all connectors.')
    for classname in self.__connections:
      try:
        self.__connections[classname].Rollback()
      except NotImplementedError:
        pass

  def PostRequest(self):
    """This cleans up any non persistent connections.
    Eg, connections that rely on request information, or connections that should
    not be kept alive beyond the scope of a request.
    """
    cleanups = [
        classname for classname in self.__connections
        if (hasattr(self.__connections[classname], 'PERSISTENT')
            and not self.__connections[classname].PERSISTENT)
    ]
    for classname in cleanups:
      try:
        self.__connections[classname].Disconnect()
      except (NotImplementedError, TypeError, ConnectionError):
        pass
      del(self.__connections[classname])

  def __iter__(self):
    """Pass tru to the Relevant connection as an Iterable, so variable unpacking
    can be used by the consuming class. This is used in the SecureCookie Model
    class."""
    return iter(self.RelevantConnection())

  def __del__(self):
    """Cleans up all references, and closes all connectors"""
    if self.debug:
      print('Deleting model connections.')
    for classname in self.__connectors:
      if not hasattr(self.__connectors[classname], 'connection'):
        continue
      try:
        self.__connections[classname].Disconnect()
      except (NotImplementedError, TypeError, ConnectionError):
        pass
